/*************************************************************************
 *  File : adj2bin.c
 *  Description: Converts adjacency graph to CSR Format. Outputs:
 *               .bin : binary form converted from input_file in ASCII
 *               .offset :binary file containing the file offsets of all
 *
 * Copyright (C) 2016, Northwestern University
 * See COPYRIGHT notice in top-level directory.
 **************************************************************************/



#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h> /* open() */
#include <sys/stat.h> /* open() */
#include <fcntl.h> /* open() */
#include <unistd.h> /* write() */
#include <errno.h>

#include <metis.h> /* need  IDXTYPEWIDTH */

#if IDXTYPEWIDTH == 32
/* typedef int idx_t; */
#define ATOI atoi
#define FMT "d"
#else
/* typedef long long idx_t; */
#define ATOI atoll
#define FMT "lld"
#endif

#define LINE_SIZE 67108864
#define CHUNK     1048576
#define DELIM     " \t\n"


/*----< main() >------------------------------------------------------------*/
int main(int argc, char **argv) {
    char *lineBuf, fname_bin[1024], fname_off[1024];
    FILE *fd1;
    int   fd2,  fd3;
    idx_t numVertices, numEdges, offset, numOff, numBin, *outBin, *outOff;
    idx_t tmpNumVertices, lineno;

    if (argc != 2) {
        fprintf(stderr,"Usage: %s file\n",argv[0]);
        exit(1);
    }

    fd1 = fopen(argv[1], "r"); /* input adjacency list file */
    if (fd1 == NULL) {
        printf("Error: open fails on file %s (%s)\n",argv[1],strerror(errno));
	exit(1);
    }
    sprintf(fname_off,"%s.offset",argv[1]); /* output offset file */
    fd2 = open(fname_off, O_CREAT|O_WRONLY|O_TRUNC, 0600);
    if (fd2 == -1) {
        printf("Error: open fails on file %s (%s)\n",fname_off,strerror(errno));
	exit(1);
    }
    sprintf(fname_bin,"%s.bin",argv[1]); /* output binary adjacency list file */
    fd3 = open(fname_bin, O_CREAT|O_WRONLY|O_TRUNC, 0600);
    if (fd3 == -1) {
        printf("Error: open fails on file %s (%s)\n",fname_bin,strerror(errno));
	exit(1);
    }

    lineBuf = (char*) malloc(LINE_SIZE);
    outBin  = (idx_t*) malloc(CHUNK * sizeof(idx_t));
    outOff  = (idx_t*) malloc(CHUNK * sizeof(idx_t));

    /* get the number of vertices and number of edges */
    if (fgets(lineBuf, LINE_SIZE, fd1) != NULL) {
        numVertices = ATOI(strtok(lineBuf, DELIM));
        numEdges    = ATOI(strtok(NULL, DELIM));
        if (strtok(NULL, DELIM) != NULL) {
            printf("Error: file format of %s is incorrect: first line can only have 2 numbers\n",argv[1]);
            exit(1);
        }
    }
    else {
        printf("Error: file %s is empty\n",argv[1]);
        exit(1);
    }
    printf("numVertices=%"FMT" numEdges=%"FMT"\n",numVertices,numEdges);
    write(fd3, &numVertices, sizeof(idx_t));
    write(fd3, &numEdges,    sizeof(idx_t));

    offset = 2 * sizeof(idx_t); /* header size */
    numBin = 0;
    outOff[0] = offset; /* first vertex starts with offset 0 */
    numOff = 1; /* calculate offsets starting from 2nd vertex */
    tmpNumVertices = 0;
    lineno = 1;
    while (fgets(lineBuf, LINE_SIZE, fd1) != NULL) {
        lineno++;
        if (lineBuf[0] == '%') continue; /* skip comment line */

        char *id = strtok(lineBuf, DELIM);
        while (id != NULL) { /* check number of tokens in the line */
            idx_t vertexID = ATOI(id);
            if (vertexID <= 0 || vertexID > numVertices) {
                printf("Error at input file line %d: vertexID %d is out of range [1: %d]\n",
                       lineno, vertexID, numVertices);
                printf("STOP ....\n");
                goto fn_exit;
            }
            offset += sizeof(idx_t);
            outBin[numBin++] = vertexID;
            if (numBin == CHUNK) { /* flush to file if buffer is full */
                write(fd3, outBin, sizeof(idx_t) * CHUNK);
                numBin = 0;
            }
            id = strtok(NULL, DELIM); /* get next token */
        }
        outOff[numOff++] = offset;

        if (numOff == CHUNK) { /* flush to file if buffer is full */
            write(fd2, outOff, sizeof(idx_t) * CHUNK);
            numOff = 0;
        }
        tmpNumVertices++;
    }
    if (numBin > 0) /* flush the remaining buffer */
        write(fd3, outBin, sizeof(idx_t) * numBin);
    if (numOff > 0) /* flush the remaining buffer */
        write(fd2, outOff, sizeof(idx_t) * numOff);

    if (tmpNumVertices != numVertices)
        printf("Warning: number of vertices read (%d) does not match with header (%d)\n",tmpNumVertices, numVertices);
    else {
        printf("Success: output binary file: %s\n",fname_bin);
        printf("Success: output offset file: %s\n",fname_off);
    }

fn_exit:
    free(lineBuf);
    free(outBin);
    free(outOff);

    fclose(fd1);
    close(fd2);
    close(fd3);

    return 0;
}
