/*************************************************************************
 * File : community_main.c (MPI version)
 * 
 * Description:  
 * In the parallel read phase the output of ParMETIS partitioner and graph
 * information are read.  The partition file contains the number of vertices
 * followed by one partition ID per line (the vertex ID is implied by the line
 * number).  The graph file contains the number of edges |E| followed by an
 * edge, which is represented as a pair of integers.  Each process reads the
 * number of vertices and edges to determined the size of each file.  The
 * partitions are stored in contiguous chunks in the file following the order
 * of process ranks.  Each process then carries out Subgraph ID Distribution,
 * Pairwise Subgraph Duplication, Local Graph Construction, Local Region
 * Growing,Local Community Merge, and Global Resolution Phase.  The community
 * assignment for each vertex is then outputed to a file 
 *
 * Author : Diana Palsetia, Wei-keng Liao, Alok Choudhary
 *          ECE Department, Northwestern University
 *          email: drp925@eecs.northwestern.edu, wkliao@eecs.northwestern.edu 
 *                 choudhar@eecs.northwestern.edu     
 * Copyright (C) 2016, Northwestern University
 * See COPYRIGHT notice in top-level directory. 
 *************************************************************************/

#include <unistd.h> /* getopt() */

#include "pmep.h"

static void
usage(char *argv0)
{
    char *help =
    "Usage: %s [-h | -v] -i input_file -k num\n"
    "       [-h] Print this help\n"
    "       [-v] verbose mode\n"
    "       input_file: input file name\n"
    "            Files input_file.bin and input_file.offset must exist\n"
    "       num: number of partitions\n"
    "            Note number of MPI processes must be num(num-1)/2\n";
    fprintf(stderr, help, argv0);
}

/*************************************************************************
* Let the game begin
**************************************************************************/
int main(int argc, char *argv[])
{
    extern int optind;
    char fname[MAX_STRING_LENGTH];
    char fname_bin[MAX_STRING_LENGTH], fname_off[MAX_STRING_LENGTH];
    double timing;
    int i, npes, mype, verbose, opt_i_presented, opt_k_presented;
    idx_t nparts, adptf, options[10], *part=NULL;
    idx_t numflag=0, wgtflag=0, edgecut, gnvtxs, gnedges;
    graph_t graph;
    real_t ipc2redist, *tpwgts=NULL, ubvec[MAXNCON];
    MPI_Comm comm=MPI_COMM_WORLD;

    MPI_Init(&argc, &argv);
    MPI_Comm_size(comm, &npes);
    MPI_Comm_rank(comm, &mype);

    /* get command-line arguments */
    verbose         = 0;
    opt_i_presented = 0;
    opt_k_presented = 0;
    while ((i = getopt(argc, argv, "hvi:k:")) != EOF) {
        switch(i) {
            case 'i': opt_i_presented = 1;
                      strcpy(fname, optarg);
                      sprintf(fname_bin, "%s.bin", optarg);
                      sprintf(fname_off, "%s.offset", optarg);
                      break;
            case 'k': opt_k_presented = 1;
                      nparts = atoi(optarg);
                      break;
            case 'v': verbose = 1;
                      break;
            case 'h':
            default:  if (mype==0) usage(argv[0]);
                      MPI_Finalize();
                      return 0;
        }
    }
    if (opt_i_presented == 0 || opt_k_presented == 0) {
        if (mype == 0) usage(argv[0]);
        MPI_Finalize();
        exit(0);
    }

    adptf      = 1; //atoi(argv[4]);
    ipc2redist = 1; //atof(argv[5]);

    options[0] = 0;
    options[PMV3_OPTION_DBGLVL] = 1; //atoi(argv[6]);
    options[PMV3_OPTION_SEED]   = 25; // atoi(argv[7]);

    MPI_Barrier(comm); /*-----------------------------------------*/
    timing = MPI_Wtime();
    ParallelReadGraph2(fname_bin, fname_off, &gnvtxs, &gnedges, &graph.nvtxs,
                       &graph.nedges, &graph.vtxdist, &graph.xadj,
                       &graph.adjncy, comm);
    double t_read = MPI_Wtime() - timing;
    MPI_Barrier(comm); /*-----------------------------------------*/
    timing = MPI_Wtime();
    wgtflag = 0;
    graph.ncon = 1;
    ubvec[0] = 1.05;
    part   = (idx_t*)  malloc(graph.nvtxs * sizeof(idx_t));
    tpwgts = (real_t*) malloc(nparts * sizeof(real_t));
    for (i=0; i<nparts; i++) tpwgts[i] = 1.0/(real_t)nparts;

    ParMETIS_V3_PartKway(graph.vtxdist, graph.xadj, graph.adjncy, NULL, NULL,
                         &wgtflag, &numflag, &graph.ncon, &nparts, tpwgts,
                         ubvec, options, &edgecut, part, &comm);
    free(tpwgts);
    double t_metis = MPI_Wtime() -timing;
    
    MPI_Barrier(comm); /*-----------------------------------------*/
    timing = MPI_Wtime();
    //Data to hold vertices for which partition ids are needed
    idx_t *dp_count = (idx_t *)calloc(npes,sizeof(idx_t));
    idx_t **dpid = (idx_t**)malloc(npes*sizeof(idx_t*));
   
    for (i=0; i< npes; i++) {
        if (i == mype)
            dpid[i] = (idx_t*)malloc(graph.nvtxs*sizeof(idx_t));
        else    
            dpid[i] = (idx_t*)malloc(((graph.nedges/2)- graph.nvtxs)*sizeof(idx_t));
    }
    idx_t hsize = (idx_t)log(graph.nedges)/log(2) + 1;
    htab * dups = hcreate(hsize); 
    exchange_partitions2_compute(dp_count,dpid,&graph,part,dups,comm);
    cleanup_table(dups);
    double t_part_compute =  MPI_Wtime() - timing;

    MPI_Barrier(comm); /*-----------------------------------------*/
    timing = MPI_Wtime();
    htab * new_part = exchange_partitions2_comm(dp_count,dpid,&graph,part,comm);
    idx_t total_exchanged = 0;
    for(i=0;i < npes;i++){
        if (i != mype)
            total_exchanged += dp_count[i];
    }
    free(dpid[0]);
    free(dpid);
    free(dp_count);
    double t_part_exchange = MPI_Wtime() - timing;

    //Create memory allocated space hold edges for each process
    //Data to store number of between edges for each process
    MPI_Barrier(comm); /*-----------------------------------------*/
    timing = MPI_Wtime();
    idx_t *bcount = (idx_t *)calloc(npes,sizeof(idx_t));
    idx_t **bedges = (idx_t **)malloc(npes*sizeof(idx_t *));
    bedges[0] = (idx_t*) malloc(npes*graph.nedges*sizeof(idx_t));
    for(i=1; i < npes;i++)
        bedges[i] = bedges[i-1] + graph.nedges;

    //Data to store number of within edges for each process
    idx_t * wcount = (idx_t *)calloc(nparts,sizeof(idx_t));
    idx_t ** wedges = (idx_t **)malloc(nparts*sizeof(idx_t *));
    wedges[0] = (idx_t *)malloc(nparts*graph.nedges*sizeof(idx_t));
    for(i=1; i < nparts;i++)
        wedges[i] = wedges[i-1] + graph.nedges;

    //Data to store edge count
    idx_t *pcount = (idx_t *)calloc(npes,sizeof(idx_t));
    idx_t **pedges = create_overlaps2(&graph,nparts,new_part,bedges,bcount,wedges,wcount,pcount,comm);

    //Free allocated memory
    free(wcount);
    free(wedges[0]);
    free(wedges);
    free(bcount);
    free(bedges[0]);
    free(bedges);
    free(part);
    free(graph.vtxdist);
    free(graph.xadj);
    free(graph.adjncy);

    cleanup_table(new_part);
    double t_overlap_commute = MPI_Wtime() - timing;

    //exchange_overlaps
    MPI_Barrier(comm); /*-----------------------------------------*/
    timing = MPI_Wtime();
    idx_t *recv_count = (idx_t *)calloc(npes,sizeof(idx_t));
    idx_t *recv_edges = exchange_overlaps(pedges,pcount,recv_count,comm);
    double t_overlap_exchange = MPI_Wtime() - timing;

    //Create graph and free list
    MPI_Barrier(comm); /*-----------------------------------------*/
    timing = MPI_Wtime();
    //Total Edges
    idx_t my_edges = 0;
    idx_t recv_total = 0;
    for(i=0;i<npes;i++)
    {
        //my_edges += recv_count[i];
        if (i != mype)
            recv_total += recv_count[i];
    }
    my_edges = pcount[mype] + recv_total;

    /* initialize edges */
    idx_t *edges = (idx_t *)malloc(my_edges * sizeof(idx_t));
    memcpy(edges,              pedges[mype], pcount[mype]*sizeof(idx_t));
    memcpy(edges+pcount[mype], recv_edges,   recv_total*sizeof(idx_t));

    //Free allocated memory space
    for(i=0; i<npes; i++)
        free(pedges[i]);
    free(pcount);
    free(recv_count);
    free(recv_edges);

    int ret;
    struct mep_graph_t mep_graph;

    idx_t direction = 1;
    ret = create_graph(my_edges, edges, direction, &mep_graph);
    if(ret != 0){
        printf("[%s][%d] build_graph failed\n", __FUNCTION__,__LINE__);
        MPI_Abort(comm, 1);
    }
    free(edges);
    double t_create_graph = MPI_Wtime() - timing;

    MPI_Barrier(comm); /*-----------------------------------------*/
    timing = MPI_Wtime();
    ret = build_free_list(&mep_graph);
    if(ret != 0){
        printf("[%s][%d] build_free_list failed\n", __FUNCTION__,__LINE__);
        MPI_Abort(comm, 1);
    }
    ret = community_grow(&mep_graph);
    if(ret != 0){
        printf("[%s][%d] community_grow failed\n", __FUNCTION__,__LINE__);
        MPI_Abort(comm, 1);
    }
    idx_t comm_b_r = hcount( (htab *)mep_graph.comm );
    double t_region_grow = MPI_Wtime() - timing;

    MPI_Barrier(comm); /*-----------------------------------------*/
    timing = MPI_Wtime();
    idx_t loops=0;
    ret = community_merge(&mep_graph, &loops);
    if(ret != 0){
        printf("[%s][%d] community_merge failed\n", __FUNCTION__,__LINE__);
        MPI_Abort( comm, 1 );
    }
    double t_merge = MPI_Wtime() - timing;

    //Exchange root and purity
    MPI_Barrier(comm); /*-----------------------------------------*/
    timing = MPI_Wtime();
    idx_t default_size,end_size,rem,o;
    idx_t * message_size = (idx_t *)calloc(npes,sizeof(idx_t));
    idx_t * offset       = (idx_t *)calloc(npes,sizeof(idx_t));

    //Determine offset and length for each process
    //Using offset and length processes know which length to work on
    default_size = gnvtxs / npes;
    rem = gnvtxs % npes;

    for(i=0;i < npes;i++)
    {
        if (rem > i)
        {
            end_size = default_size + 1;
            message_size[i] = end_size;
            o = end_size * i;
            offset[i] = o;
        }
        else
        {
            end_size = default_size;
            message_size[i] = end_size;
            o = end_size*i+rem;
            offset[i] = o;
        }
    }

    idx_t * send_root_purity =(idx_t *)calloc(2*gnvtxs,sizeof(idx_t));
    unsigned int j;
//    int ret;
    struct vertex_t *vertex=NULL;
    for(j=0; j< mep_graph.num_vertex; j++){
        ret = hfind((htab *)mep_graph.adj, &mep_graph.vertices[j], sizeof(idx_t));
        if(ret == FALSE){
            printf("[%s][%d] hfind failed\n", __FUNCTION__,__LINE__);
            MPI_Abort( comm, 1 );
        }
        vertex = hstuff((htab *)mep_graph.adj);
        //printf("%u    %u %u\n", vertex->key, find(&mep_graph, vertex),vertex->comp_max);
        send_root_purity[2*(vertex->key-1)] = vertex->root;//find(&mep_graph, vertex);
        send_root_purity[2*(vertex->key-1)+1] = vertex->comp_max;
    }

    idx_t * recv_root_purity =  exchange_root_purity(send_root_purity,gnvtxs,message_size,offset,comm);
    double t_exchange_root_purity = MPI_Wtime() - timing;

    //resolve
    MPI_Barrier(comm); /*-----------------------------------------*/
    timing = MPI_Wtime();
    idx_t * final_root_purity = resolve(recv_root_purity, send_root_purity,message_size,offset,npes,mype);
    free(send_root_purity);
    free(recv_root_purity);
    double t_resolve = MPI_Wtime() - timing;

    //gather final root
    MPI_Barrier(comm); /*-----------------------------------------*/
    timing = MPI_Wtime();
    idx_t * final_root  = gather_root(message_size,offset,final_root_purity,gnvtxs,comm);
    free(message_size);
    free(offset);
    free(final_root_purity);
    double t_gather = MPI_Wtime() - timing;

    //output the final communities
    MPI_Barrier(comm); /*-----------------------------------------*/
    timing = MPI_Wtime();
    write_community(fname, gnvtxs, final_root, comm);
    if(mype == 0)
        free(final_root);

    //Delete graph
    delete_graph(&mep_graph);
    double t_write = MPI_Wtime() - timing;

    idx_t total_o = 0.0;
    unsigned int total_e = 0;
    idx_t total_e_per_p_max = 0;
    idx_t total_e_per_p_min = 0;

    idx_t total_part_exchange = 0;
    idx_t max_part_exchange = 0;
    idx_t min_part_exchange = 0;
    
    idx_t sum_comm_before_merge = 0;
    idx_t max_comm_before_merge = 0;
    idx_t min_comm_before_merge = 0;
    idx_t m_loops_sum = 0;
    idx_t m_loops_max = 0;
    idx_t m_loops_min = 0;
 
    double m_t_read = 0.0;
    double m_t_metis = 0.0;
    double m_t_part_compute = 0.0;
    double m_t_part_exchange = 0.0;
    double m_t_overlap_commute = 0.0;
    double m_t_overlap_exchange  = 0.0;
    double m_t_create_graph = 0.0;
    double m_t_region_grow  = 0.0;
    double m_t_merge = 0.0;
    double m_t_exchange_root_purity = 0.0;
    double m_t_resolve = 0.0;
    double m_t_gather = 0.0;
    double m_t_write = 0.0;
    double total = 0.0;
    double total_w_write = 0.0;

    total_o = gnedges;
    MPI_Reduce(&my_edges, &total_e, 1, MPI_UNSIGNED,MPI_SUM,0,comm);
    MPI_Reduce(&my_edges, &total_e_per_p_max, 1, IDX_T,MPI_MAX,0,comm);
    MPI_Reduce(&my_edges, &total_e_per_p_min, 1, IDX_T,MPI_MIN,0,comm);

    MPI_Reduce(&total_exchanged, &total_part_exchange, 1, IDX_T,MPI_SUM,0,comm);
    MPI_Reduce(&total_exchanged, &max_part_exchange, 1, IDX_T,MPI_MAX,0,comm);
    MPI_Reduce(&total_exchanged, &min_part_exchange, 1, IDX_T,MPI_MIN,0,comm);
    
    MPI_Reduce(&comm_b_r, &sum_comm_before_merge, 1, IDX_T,MPI_SUM,0,comm);
    MPI_Reduce(&comm_b_r, &max_comm_before_merge, 1, IDX_T,MPI_MAX,0,comm);
    MPI_Reduce(&comm_b_r, &min_comm_before_merge, 1, IDX_T,MPI_MIN,0,comm);
    
    
    MPI_Reduce(&loops, &m_loops_sum, 1, IDX_T,MPI_SUM,0,comm);
    MPI_Reduce(&loops, &m_loops_max, 1, IDX_T,MPI_MAX,0,comm);
    MPI_Reduce(&loops, &m_loops_min, 1, IDX_T,MPI_MIN,0,comm);
    
    MPI_Reduce(&t_read, &m_t_read, 1, MPI_DOUBLE,MPI_MAX,0,comm);
    MPI_Reduce(&t_metis, &m_t_metis, 1, MPI_DOUBLE,MPI_MAX,0,comm);
    MPI_Reduce(&t_part_compute, &m_t_part_compute, 1, MPI_DOUBLE,MPI_MAX,0,comm);
    MPI_Reduce(&t_part_exchange, &m_t_part_exchange, 1, MPI_DOUBLE,MPI_MAX,0,comm);
    MPI_Reduce(&t_overlap_commute, &m_t_overlap_commute, 1, MPI_DOUBLE,MPI_MAX,0,comm);
    MPI_Reduce(&t_overlap_exchange, &m_t_overlap_exchange, 1, MPI_DOUBLE,MPI_MAX,0,comm);
    MPI_Reduce(&t_create_graph, &m_t_create_graph, 1, MPI_DOUBLE,MPI_MAX,0,comm);
    MPI_Reduce(&t_region_grow, &m_t_region_grow, 1, MPI_DOUBLE,MPI_MAX,0,comm);
    MPI_Reduce(&t_merge, &m_t_merge, 1, MPI_DOUBLE,MPI_MAX,0,comm);
    MPI_Reduce(&t_exchange_root_purity, &m_t_exchange_root_purity, 1, MPI_DOUBLE,MPI_MAX,0,comm);
    MPI_Reduce(&t_resolve, &m_t_resolve, 1, MPI_DOUBLE,MPI_MAX,0,comm);
    MPI_Reduce(&t_gather, &m_t_gather, 1, MPI_DOUBLE,MPI_MAX,0,comm);
    MPI_Reduce(&t_write, &m_t_write, 1, MPI_DOUBLE,MPI_MAX,0,comm);

    if (mype == 0) {
        char partfile[256];
        sprintf(partfile, "%s_comm%"PRIDX".txt",fname,npes);

        if (verbose) {
            char partfile[256];
            printf("total number of processes                %12"PRIDX"\n",npes);
            printf("total edges in the graph                        %12"PRIDX"\n",total_o/2);
            printf("total edges due to pairwise duplication         %12"PRIDX"\n",total_e/2);
            printf("max no. of edges due to pairwise duplication    %12"PRIDX"\n",total_e_per_p_max/2);
            printf("min no. of edges due to pairwise duplication    %12"PRIDX"\n",total_e_per_p_min/2);
            printf("total vertices exchanged in id distribution     %12"PRIDX"\n",total_part_exchange);
            printf("max no. of vertices exchanged in id distribution%12"PRIDX"\n",max_part_exchange);
            printf("min no. of vertices exchanged in id distribution%12"PRIDX"\n",min_part_exchange);
            printf("total communities before merge                  %12"PRIDX"\n",sum_comm_before_merge);
            printf("max no. of communities before merge             %12"PRIDX"\n",max_comm_before_merge);
            printf("min no. of communities before merge             %12"PRIDX"\n",min_comm_before_merge);        
            printf("total no. of merge loop iterations              %12"PRIDX"\n",m_loops_sum);
            printf("max no. of merge loop iterations                %12"PRIDX"\n",m_loops_max);
            printf("min no. of merge loop iterations                %12"PRIDX"\n",m_loops_min);

            printf("----------------------TIME BREAKDOWN--------------------------------\n");

            printf("Time in parallel read                  %10.4f sec\n",m_t_read);
            printf("Time in parallel metis                 %10.4f sec\n",m_t_metis);
            printf("Time in id distribution - compute      %10.4f sec\n",m_t_part_compute);
            printf("Time in id distribution - exchange     %10.4f sec\n",m_t_part_exchange);
            printf("Time in pairwise duplication - compute %10.4f sec\n",m_t_overlap_commute);
            printf("Time in pairwise duplication - exchange%10.4f sec\n",m_t_overlap_exchange);
            printf("Time in create_graph                   %10.4f sec\n",m_t_create_graph);
            printf("Time in region_grow                    %10.4f sec\n",m_t_region_grow);
            printf("Time in merge                          %10.4f sec\n",m_t_merge);
            printf("Time in exchange_root_purity           %10.4f sec\n",m_t_exchange_root_purity);
            printf("Time in resolve                        %10.4f sec\n",m_t_resolve);
            total = m_t_read + m_t_metis + m_t_part_compute + m_t_part_exchange + m_t_overlap_commute + m_t_overlap_exchange + t_create_graph + t_region_grow + t_merge + m_t_exchange_root_purity+ t_resolve;
            printf("Time total                             %10.4f sec\n",total);
            total_w_write = total + m_t_gather + m_t_write;
            printf("Time total + write                     %10.4f sec\n",total_w_write);
        }
        else{
            printf("total number of processes                       %12"PRIDX"\n",npes);
            printf("total edges in the graph                        %12"PRIDX"\n",total_o/2);
            printf("total edges due to pairwise duplication         %12"PRIDX"\n",total_e/2);
            printf("max no. of edges due to pairwise duplication    %12"PRIDX"\n",total_e_per_p_max/2);
            printf("min no. of edges due to pairwise duplication    %12"PRIDX"\n",total_e_per_p_min/2);
            printf("total vertices exchanged in id distribution     %12"PRIDX"\n",total_part_exchange);
            printf("max no. of vertices exchanged in id distribution%12"PRIDX"\n",max_part_exchange);
            printf("min no. of vertices exchanged in id distribution%12"PRIDX"\n",min_part_exchange);
            printf("total communities before merge                  %12"PRIDX"\n",sum_comm_before_merge);
            printf("max no. of communities before merge             %12"PRIDX"\n",max_comm_before_merge);
            printf("min no. of communities before merge             %12"PRIDX"\n",min_comm_before_merge);
            printf("total no. of merge loop iterations              %12"PRIDX"\n",m_loops_sum);
            printf("max no. of merge loop iterations                %12"PRIDX"\n",m_loops_max);
            printf("min no. of merge loop iterations                %12"PRIDX"\n",m_loops_min);
            printf("----------------------TIMING RESULTS --------------------------------\n");
            total = m_t_read + m_t_metis + m_t_part_compute + m_t_part_exchange + m_t_overlap_commute + m_t_overlap_exchange + t_create_graph + t_region_grow + t_merge + m_t_exchange_root_purity+ t_resolve;
            printf("Time total                             %10.4f sec\n",total);
            total_w_write = total + m_t_gather + m_t_write;
            printf("Time total + write                     %10.4f sec\n",total_w_write);
        } 
    }

    MPI_Finalize();

    return 0;
}



/*************************************************************************
* This function changes the numbering to be from 1 instead of 0
**************************************************************************/
void ChangeToFortranNumbering(idx_t *vtxdist, idx_t *xadj, idx_t *adjncy, idx_t mype, idx_t npes)
{
  idx_t i, nvtxs, nedges;

  nvtxs = vtxdist[mype+1]-vtxdist[mype];
  nedges = xadj[nvtxs];

  for (i=0; i<npes+1; i++)
    vtxdist[i]++;
  for (i=0; i<nvtxs+1; i++)
    xadj[i]++;
  for (i=0; i<nedges; i++)
    adjncy[i]++;

  return;
}
