/*
 *
 * This file contains routines related to mep algorithm
 *
 * Diana Palsetia
 * 
 * Copyright (C) 2016, Northwestern University
 * See COPYRIGHT notice in top-level directory.
 *
 */
 
#include <stdio.h>
#include <sys/time.h>
#include <malloc.h>
#include <string.h>
#include <math.h>
#include "pmep.h"

#define DEBUG 0

#if DEBUG
static double wtime(void)
{
    double          now_time;
    struct timeval  etstart;
    struct timezone tzp;

    if (gettimeofday(&etstart, &tzp) == -1)
        perror("Error: calling gettimeofday() not successful.\n");

    now_time = ((double)etstart.tv_sec) + ((double)etstart.tv_usec) / 1000000.0;
    return now_time;
}
#endif

idx_t find(struct mep_graph_t *graph, struct vertex_t *vertex)
{
	int ret=0;
	struct vertex_t *cur=vertex;

	while(cur->root != cur->key){
		ret = hfind((htab *)graph->adj, &cur->root, sizeof(idx_t));
		if(ret == FALSE){
			printf("[%s][%d] hfind failed\n", __FUNCTION__,__LINE__);
			return -1;
		}
		cur = (struct vertex_t *)hstuff((htab *)graph->adj);
	}
	return cur->key;
}

static int update_comm(struct mep_graph_t *graph, struct vertex_t *vertex, struct vertex_t *parent)
{
	int ret=0;
	idx_t top=0;
	idx_t parent_top=0;
	struct vertex_t *root=NULL;
	struct vertex_t *parent_root=NULL;

	top = find(graph, vertex);
	ret = hfind((htab *)graph->adj, &top, sizeof(idx_t));
	if(ret != TRUE){
		printf("[%s][%d] something wrong... %u vertex is not in adj table\n", __FUNCTION__,__LINE__, top);
		return -1;
	}
	root = hstuff((htab *)graph->adj);
	
	parent_top = find(graph, parent);
	ret = hfind((htab *)graph->adj, &parent_top, sizeof(idx_t));
	if(ret != TRUE){
		printf("[%s][%d] something wrong... %u vertex is not in adj table\n", __FUNCTION__,__LINE__, top);
		return -1;
	}
	parent_root = hstuff((htab *)graph->adj);

	/* sometimes the new root is a free vertex. In that case, the new root should be added into comm table. */
	ret = hfind((htab *)graph->comm, &top, sizeof(idx_t));
	if(ret == FALSE){
		if(root->key != root->root){
			printf("[%s][%d] find function returned the wrong root vertex\n", __FUNCTION__,__LINE__);
			return -1;
		}

		if(root->free == 1)
			root->free = 0;

		ret = hadd((htab *)graph->comm, &root->key, sizeof(idx_t), (void *)root);
		if(ret == FALSE){
			printf("[%s][%d] hadd failed\n", __FUNCTION__,__LINE__);
			return -1;
		}
	}

	ret = hfind((htab *)graph->comm, &vertex->key, sizeof(idx_t));
	if(ret == TRUE){
		hdel((htab *)graph->comm);
	}
	return 0;
}

static int update_sep(struct mep_graph_t *graph, struct vertex_t *vertex, struct vertex_t *root, int debug)
{
	int ret=0;
	unsigned int i=0;
	idx_t community;
	idx_t *key=NULL;
	float *value=NULL;
	struct vertex_t *neighbor=NULL;

	for(i=0; i<vertex->degree; i++){
		ret = hfind((htab *)graph->adj, &vertex->neighbors[i], sizeof(idx_t));
		if(ret == FALSE){
			printf("[%s][%d] hfind failed\n", __FUNCTION__,__LINE__);
			return -1;
		}
		neighbor = hstuff((htab *)graph->adj);
		community = find(graph, neighbor);
		
		if(community == root->key){
			root->compactness += 1.0;
		}
		else{
/* update root sep table */
			ret = hfind((htab *)root->sep, &community, sizeof(idx_t));
			if(ret == FALSE){
			/* add a new entry in this Cluster */
				key = (idx_t *)malloc(sizeof(idx_t));
				if(key == NULL){
					printf("[%s][%d] out of memory\n", __FUNCTION__,__LINE__);
					return -1;
				}
				value = (float *)malloc(sizeof(float));
				if(key == NULL){
					printf("[%s][%d] out of memory\n", __FUNCTION__,__LINE__);
					return -1;
				}

				*key = community;
				*value = (float)1.0;
				ret = hadd((htab *)root->sep, key, sizeof(idx_t), (void *)value);
				if(ret == FALSE){
					printf("[%s][%d] hadd failed\n", __FUNCTION__,__LINE__);
					return -1;
				}
			}
			else{
				value = (float *)hstuff((htab *)root->sep);
				*value = (*value) + (float)1.0;
			}
		}
	}
	return 0;
}

static int update_community(struct mep_graph_t *graph, struct vertex_t *root, int debug)
{
	int ret=0;
	unsigned int order=0;
	unsigned int num_direct_clusters=0;
	idx_t *key=NULL;
	float *value=NULL;
	struct vertex_t *vertex=NULL;

	order = log((double)graph->num_comm);

/* 1. remove the current value */
	root->compactness = 0;
	root->avg_sep = 0;
	if(root->sep != NULL){
		while(hcount((htab *)root->sep)){
			hfirst((htab *)root->sep);
			key = (idx_t *)hkey((htab *)root->sep);
			if(key != NULL){
				free(key);
				key = NULL;
			}
			value = (float *)hstuff((htab *)root->sep);
			if(value != NULL){
				free(value);
				value = NULL;
			}
			hdel((htab *)root->sep);
		}
		hdestroy((htab *)root->sep);
	}
	root->sep = hcreate((idx_t)order+1);
	vertex = root;

/* 2. update the new value */
	while(vertex->cycle != root->key){
		ret = update_sep(graph, vertex, root, debug);
		if(ret != 0){
			printf("[%s][%d] update_sep failed\n", __FUNCTION__,__LINE__);
			return -1;
		}
		
		ret = hfind((htab *)graph->adj, &vertex->cycle, sizeof(idx_t));
		if(ret == FALSE){
			printf("[%s][%d] hfind failed\n", __FUNCTION__,__LINE__);
			return -1;
		}
		vertex = hstuff((htab *)graph->adj);
	}

	ret = update_sep(graph, vertex, root, debug);
	if(ret != 0){
		printf("[%s][%d] update_sep failed\n", __FUNCTION__,__LINE__);
		return -1;
	}

/* 3. calculate the compactness & average separation */
	root->compactness = root->compactness/(float)2.0;

	num_direct_clusters = 0;
	if(hfirst((htab *)root->sep)){
		do{
			key = (idx_t *)hkey((htab *)root->sep);
			value = (float *)hstuff((htab *)root->sep);

			if((*value != 0)&&(*key != root->key)){
				num_direct_clusters++;
			}
			root->avg_sep += (float)*value;
		}while(hnext((htab *)root->sep));
	}

	if(num_direct_clusters == 0){
		root->avg_sep = 0;
	}
	else{
		root->avg_sep /= (float)num_direct_clusters;
	}

	return 0;
}

int find_pure_community(struct mep_graph_t *graph, struct vertex_t *vertex, idx_t *target, unsigned int *found)
{
	int hret=0;
	idx_t max_key=0;
	idx_t neighbor_root_key=0;
	unsigned int i=0;
	unsigned int order=0;
	unsigned int num_free=0;
	unsigned int comp_max=0;
	htab *neighbor_table=NULL;
	struct vertex_t *neighbor=NULL;
	struct vertex_t *neighbor_root=NULL;

	order = log((double)vertex->degree);
	neighbor_table = hcreate(order);
	
	for(i=0; i<vertex->degree; i++){
		hret = hfind((htab *)graph->adj, &vertex->neighbors[i], sizeof(idx_t));
		if(hret == FALSE){
			printf("[%s][%d] hfind failed\n", __FUNCTION__,__LINE__);
			return -1;
		}
		neighbor = hstuff((htab *)graph->adj);
		
		if(neighbor->free == 1){
			num_free++;
			continue;
		}
		
		neighbor_root_key = find(graph, neighbor);
		
		hret = hfind((htab *)graph->adj, &neighbor_root_key, sizeof(idx_t));
		if(hret == FALSE){
			printf("[%s][%d] hfind failed\n", __FUNCTION__,__LINE__);
			return -1;
		}
		neighbor_root = hstuff((htab *)graph->adj);
		
		hret = hfind((htab *)neighbor_table, &neighbor_root->key, sizeof(idx_t));
		if(hret == FALSE){
			neighbor_root->comp = 1;
			hret = hadd((htab *)neighbor_table, &neighbor_root->key, sizeof(idx_t), (void *)neighbor_root);
			if(hret == FALSE){
				printf("[%s][%d] hadd failed\n", __FUNCTION__,__LINE__);
				return -1;
			}
		}
		else{
			neighbor_root = hstuff((htab *)neighbor_table);
			neighbor_root->comp++;
		}
	}

	unsigned int all_the_same=1;
	unsigned int previous_comp=0;
	idx_t min_key=0;
	
	if(hfirst((htab *)neighbor_table)){
		neighbor_root = hstuff((htab *)neighbor_table);
		previous_comp = neighbor_root->comp;
		min_key = neighbor_root->key;

		do{
			neighbor_root = hstuff((htab *)neighbor_table);
			if(neighbor_root->comp != previous_comp)
				all_the_same = 0;
			if(min_key > neighbor_root->key)
				min_key = neighbor_root->key;

			if(comp_max < neighbor_root->comp){
				max_key = neighbor_root->key;
				comp_max = neighbor_root->comp;
			}	
		}while(hnext((htab *)neighbor_table));
	}
	
	if(all_the_same){
		max_key = min_key;
	}

	if(num_free <= comp_max){
		*found = 1;
		*target = max_key;
		vertex->comp_max = comp_max;
	}
	else{
		*found = 0;
		vertex->comp_max = 0;
	}
	
	while(hcount((htab *)neighbor_table)){
		hfirst((htab *)neighbor_table);
		hdel((htab *)neighbor_table);
	}
	hdestroy(neighbor_table);
	
	return 0;
}

int find_pure_community_new(struct mep_graph_t *graph, struct vertex_t *vertex, idx_t *target, unsigned int *found)
{
	int hret=0;
	idx_t max_key=0;
	idx_t neighbor_root_key=0;
	unsigned int i=0;
	unsigned int order=0;
	unsigned int num_free=0;
	unsigned int comp_max=0;
	htab *neighbor_table=NULL;
	struct vertex_t *neighbor=NULL;
	struct vertex_t *neighbor_root=NULL;

	order = log((double)vertex->degree);
	neighbor_table = hcreate(order);
	
	for(i=0; i<vertex->degree; i++){
		hret = hfind((htab *)graph->adj, &vertex->neighbors[i], sizeof(idx_t));
		if(hret == FALSE){
			printf("[%s][%d] hfind failed\n", __FUNCTION__,__LINE__);
			return -1;
		}
		neighbor = hstuff((htab *)graph->adj);
		
		if(neighbor->free == 1){
			printf("Found a free neighbor\n");
			num_free++;
			continue;
		}
		
		neighbor_root_key = find(graph, neighbor);
		
		hret = hfind((htab *)graph->adj, &neighbor_root_key, sizeof(idx_t));
		if(hret == FALSE){
			printf("[%s][%d] hfind failed\n", __FUNCTION__,__LINE__);
			return -1;
		}
		neighbor_root = hstuff((htab *)graph->adj);
		
		hret = hfind((htab *)neighbor_table, &neighbor_root->key, sizeof(idx_t));
		if(hret == FALSE){
			neighbor_root->comp = 1;
			hret = hadd((htab *)neighbor_table, &neighbor_root->key, sizeof(idx_t), (void *)neighbor_root);
			if(hret == FALSE){
				printf("[%s][%d] hadd failed\n", __FUNCTION__,__LINE__);
				return -1;
			}
		}
		else{
			neighbor_root = hstuff((htab *)neighbor_table);
			neighbor_root->comp++;
		}
	}

	unsigned int all_the_same=1;
	unsigned int previous_comp=0;
	idx_t min_key=0;
	
	if(hfirst((htab *)neighbor_table)){
		neighbor_root = hstuff((htab *)neighbor_table);
		previous_comp = neighbor_root->comp;
		min_key = neighbor_root->key;

		do{
			neighbor_root = hstuff((htab *)neighbor_table);
			if(neighbor_root->comp != previous_comp)
				all_the_same = 0;
			if(min_key > neighbor_root->key)
				min_key = neighbor_root->key;

			if(comp_max < neighbor_root->comp){
				max_key = neighbor_root->key;
				comp_max = neighbor_root->comp;
			}	
		}while(hnext((htab *)neighbor_table));
	}
	
	if(all_the_same){
		max_key = min_key;
	}

	if(num_free <= comp_max){
		*found = 1;
		*target = max_key;
		vertex->comp_max += comp_max;
	}
	else{
		*found = 0;
		vertex->comp_max = 0;
	}
	
	while(hcount((htab *)neighbor_table)){
		hfirst((htab *)neighbor_table);
		hdel((htab *)neighbor_table);
	}
	hdestroy(neighbor_table);
	
	return 0;
}


//int find_community2(struct mep_graph_t *graph, struct vertex_t *vertex, idx_t *target, idx_t *found, idx_t *comp_max);

int find_community2(struct mep_graph_t *graph, struct vertex_t *vertex, idx_t *target, idx_t *found, idx_t *comp_max)
{
	idx_t i=0;
	int ret=0;
	unsigned int num_free=0;
	unsigned int num_communities;
	idx_t community=0;
	idx_t max_value=0;
	idx_t *communities=NULL;
	struct vertex_t *neighbor=NULL;
	idx_t *count=NULL;

	if(graph == NULL || vertex == NULL || target == NULL || found == NULL){
		printf("[%s][%d] inavalid parameter\n", __FUNCTION__,__LINE__);
		return -1;
	}

	communities = (idx_t *)malloc(sizeof(idx_t)*vertex->degree);
	if(communities == NULL){
		printf("[%s][%d] out of memory\n", __FUNCTION__,__LINE__);
		return -1;
	}

/* 1. calculate the number of free neighbors */
	for(i=0; i<vertex->degree; i++){
		ret = hfind((htab *)graph->adj, &vertex->neighbors[i], sizeof(idx_t));
		if(ret == FALSE){
			printf("[%s][%d] hfind failed\n", __FUNCTION__,__LINE__);
			return -1;
		}
		neighbor = (struct vertex_t *)hstuff((htab *)graph->adj);	
		num_free += neighbor->free;
		communities[i] = find(graph, neighbor);
		if(max_value < communities[i]){
			max_value = communities[i];
		}
	}
	num_communities = max_value;

/* 2. Create a new community if the free neighbors are more than half of degree */
	if(num_free > (unsigned int)(vertex->degree*0.5)){
		*found = 0;
		free(communities);
		return 0;
	}

/* 3. Find the community with most edges */
	count = (idx_t *)malloc(sizeof(idx_t)*(max_value+1));
	if(count == NULL){
		printf("[%s][%d] out of memory\n", __FUNCTION__,__LINE__);
		return -1;
	}
	memset(count, 0, sizeof(idx_t)*(max_value+1));

	/* count[1 ~ max_value] is only valid... because clusters[i] starts from 1, not 0. */
	for(i=0; i<vertex->degree; i++){
		if(communities[i] != 0)
			count[communities[i]]++;
	}

	max_value = count[0];
	community = 0;
	for(i=1; i<(num_communities+1); i++){
		if(max_value < count[i]){
			max_value = count[i];
			community = i;
		}
	}

	if(community == 0){	
		printf("[%s][%d] comm is 0... something wrong\n", __FUNCTION__,__LINE__);
		free(count);
		free(communities);
		return -1;
	}

	vertex->comp_max = max_value;
	*comp_max = max_value;

	if(num_free <= max_value){
		*target = community;
		*found = 1;
	}
	else{
		*found = 0;
	}

	if(count != NULL)
		free(count);
	if(communities != NULL)
		free(communities);
	return 0;
}

static int union_remsp(struct mep_graph_t *graph, struct vertex_t *vertex_x, struct vertex_t *vertex_y)
{
	int ret=0;
	idx_t temp_key=0;

	while(vertex_x->root != vertex_y->root){
		if(vertex_x->root > vertex_y->root){
			if(vertex_x->root == vertex_x->key){
				vertex_x->root = vertex_y->root;

/* Community Root is merged... remove vertex_x from the comm hash table */
				ret = update_comm(graph, vertex_x, vertex_y);
				if(ret != 0){
					printf("[%s][%d] update_comm failed\n", __FUNCTION__,__LINE__);
					return -1;
				}

				return 0;
			}
			temp_key = vertex_x->root;
			vertex_x->root = vertex_y->root;

			ret = hfind((htab *)graph->adj, &temp_key, sizeof(idx_t));
			if(ret == FALSE){
				printf("[%s][%d] hfind failed\n", __FUNCTION__,__LINE__);
				return -1;
			}
			vertex_x = (struct vertex_t *)hstuff((htab *)graph->adj);
		}
		else{
			if(vertex_y->root == vertex_y->key){
				vertex_y->root = vertex_x->root;

/* Community Root is merged... remove vertex_y from the comm hash table */
				ret = update_comm(graph, vertex_y, vertex_x);
				if(ret != 0){
					printf("[%s][%d] update_comm failed\n", __FUNCTION__,__LINE__);
					return -1;
				}
				return 0;
			}
			temp_key = vertex_y->root;
			vertex_y->root = vertex_x->root;
		
			ret = hfind((htab *)graph->adj, &temp_key, sizeof(idx_t));
			if(ret == FALSE){
				printf("[%s][%d] hfind failed\n", __FUNCTION__,__LINE__);
				return -1;
			}
			vertex_y = (struct vertex_t *)hstuff((htab *)graph->adj);
		}
	}
	return 0;
}

static int union_cycle(struct mep_graph_t *graph, struct vertex_t *vertex_x, struct vertex_t *vertex_y)
{
	idx_t cycle_x=0;
	idx_t cycle_y=0;

	cycle_x = vertex_x->cycle;
	cycle_y = vertex_y->cycle;
	vertex_x->cycle = cycle_y;
	vertex_y->cycle = cycle_x;

	return 0;
}

static int remove_root(struct mep_graph_t *graph, struct vertex_t *vertex)
{
	int ret=0;
	idx_t new_root_key=0;
	struct vertex_t *current=NULL;
	
/* 1. find a new root vertex */
	if(vertex->root == vertex->key){/* removing root vertex */
		/* find the next minimum root key */
		new_root_key = vertex->cycle;
		ret = hfind((htab *)graph->adj, &new_root_key, sizeof(idx_t));
		if(ret == FALSE){
			printf("[%s][%d] hfind failed\n", __FUNCTION__,__LINE__);
			return -1;
		}
		current = hstuff((htab *)graph->adj);
		while(1){
			if(new_root_key > current->key){
				new_root_key = current->key;
			}
			
			if(current->cycle == vertex->key)
				break;

			ret = hfind((htab *)graph->adj, &current->cycle, sizeof(idx_t));
			if(ret == FALSE){
				printf("[%s][%d] hfind failed\n", __FUNCTION__,__LINE__);
				return -1;
			}
			current = hstuff((htab *)graph->adj);
		}

		ret = hfind((htab *)graph->adj, &new_root_key, sizeof(idx_t));
		if(ret == FALSE){
			printf("[%s][%d] hfind failed\n", __FUNCTION__,__LINE__);
			return -1;
		}
		current = hstuff((htab *)graph->adj);
		current->root = current->key;

		/* replace the community with the new one in comm hash table */
		ret = hfind((htab *)graph->comm, &vertex->key, sizeof(idx_t));
		if(ret == FALSE){
			printf("[%s][%d] hfind failed\n", __FUNCTION__,__LINE__);
			return -1;
		}
		hdel((htab *)graph->comm);

		ret = hadd((htab *)graph->comm, &current->key, sizeof(idx_t), (void *)current);
		if(ret == FALSE){
			printf("[%s][%d] hadd failed\n", __FUNCTION__,__LINE__);
			return -1;
		}
	}
	else{
		new_root_key = vertex->root;
		ret = hfind((htab *)graph->adj, &new_root_key, sizeof(idx_t));
		if(ret == FALSE){
			printf("[%s][%d] hfind failed\n", __FUNCTION__,__LINE__);
			return -1;
		}
		current = hstuff((htab *)graph->adj);
	}

/* 2. traverse the community and replace the root value with the new root */
	while(1){
		ret = hfind((htab *)graph->adj, &current->cycle, sizeof(idx_t));
		if(ret == FALSE){
			printf("[%s][%d] hfind failed\n", __FUNCTION__,__LINE__);
			return -1;
		}
		current = hstuff((htab *)graph->adj);
		if(current->key == new_root_key)
			break;
		
		if(current->root == vertex->key)
			current->root = new_root_key;
	}

	vertex->root = vertex->key;
	return 0;
}

static int remove_cycle(struct mep_graph_t *graph, struct vertex_t *vertex)
{
	int ret=0;
	struct vertex_t *vertex_before=NULL;

/* 1. find a vertex which indicates the x vertex as a cycle */
	vertex_before = vertex;

	do{
		ret = hfind((htab *)graph->adj, &vertex_before->cycle, sizeof(idx_t));
		if(ret == FALSE){
			printf("[%s][%d] hfind failed\n", __FUNCTION__,__LINE__);
			return -1;
		}
		vertex_before = (struct vertex_t *)hstuff((htab *)graph->adj);
	}while(vertex_before->cycle != vertex->key);
	
	
/* 2. remove the x vertex */
	vertex_before->cycle = vertex->cycle;
	vertex->cycle = vertex->key;
	return 0;
}

static int update_neighbors(struct mep_graph_t *graph, struct vertex_t *vertex)
{
	unsigned int i=0;
	int ret=0;
	unsigned int found=0;
	idx_t target=0;
	idx_t community=0;
	struct vertex_t *neighbor;

	community = find(graph, vertex);

	for(i=0; i<vertex->degree; i++){
		ret = hfind((htab *)graph->adj, &vertex->neighbors[i], sizeof(idx_t));
		if(ret == FALSE){
			printf("[%s][%d] hfind failed\n", __FUNCTION__,__LINE__);
			return -1;
		}
		neighbor = (struct vertex_t *)hstuff((htab *)graph->adj);
		if(neighbor->free == 1){
			ret = find_pure_community(graph, neighbor, &target, &found);
			if(ret != 0){
				printf("[%s][%d] find_pure_community failed\n", __FUNCTION__,__LINE__);
				return -1;
			}
			
			if(found){
				if(community == target){
					neighbor->free = 0;

					ret = union_remsp(graph, vertex, neighbor);
					if(ret != 0){
						printf("[%s][%d] union_remsp failed\n", __FUNCTION__,__LINE__);
						return -1;
					}

					ret = union_cycle(graph, vertex, neighbor);
					if(ret != 0){
						printf("[%s][%d] union_cycle failed\n", __FUNCTION__,__LINE__);
						return -1;
					}

					ret = update_neighbors(graph, neighbor);
					if(ret != 0){
						printf("[%s][%d] update_neighbors failed\n", __FUNCTION__,__LINE__);
						return -1;
					}
				}
			}
		}
	}
	return 0;
}

/**************************************************************************
 * What about ignoring neighbors here and calling this function several times?
 * It would be more easy to be parallelized...
**************************************************************************/
static int update_vertex(struct mep_graph_t *graph, struct vertex_t *vertex)
{
	unsigned int i=0;
	int ret=0;
	unsigned int found=0;
	idx_t target=0;
	idx_t root=0;
	idx_t neighbor_root=0;
	struct vertex_t *neighbor=NULL;
	struct vertex_t *new_root=NULL;

/* 1. find the purest community */
	ret = find_pure_community(graph, vertex, &target, &found);
	if(ret != 0){
		printf("[%s][%d] find_pure_community failed\n", __FUNCTION__,__LINE__);
		return -1;
	}

	if(found){
		ret = hfind((htab *)graph->adj, &target, sizeof(idx_t));
		if(ret == FALSE){
			printf("[%s][%d] hfind failed\n", __FUNCTION__,__LINE__);
			return -1;
		}
		new_root = hstuff((htab *)graph->adj);

		vertex->free = 0;
		ret = union_remsp(graph, new_root, vertex);
		if(ret != 0){
			printf("[%s][%d] union_remsp failed\n", __FUNCTION__,__LINE__);
			return -1;
		}

		ret = union_cycle(graph, new_root, vertex);
		if(ret != 0){
			printf("[%s][%d] union_cycle failed\n", __FUNCTION__,__LINE__);
			return -1;
		}
	}
	else{
/* 2. create a new cluster */
		vertex->free = 0;
		vertex->root = vertex->key;

		ret = hadd((htab *)graph->comm, &vertex->key, sizeof(idx_t), (void *)vertex);
		if(ret == FALSE){
			printf("[%s][%d] hadd failed\n", __FUNCTION__,__LINE__);
			return -1;
		}
		graph->num_comm++;
		
/* 3. merge all the free & direct neighbors */
		for(i=0; i<vertex->degree; i++){
			ret = hfind((htab *)graph->adj, &vertex->neighbors[i], sizeof(idx_t));
			if(ret == FALSE){
				printf("[%s][%d] hfind failed\n", __FUNCTION__,__LINE__);
				return -1;
			}
			neighbor = (struct vertex_t *)hstuff((htab *)graph->adj);
			if(neighbor->free == 1){
				neighbor->free = 0;

				ret = union_remsp(graph, vertex, neighbor);
				if(ret != 0){
					printf("[%s][%d] union_remsp is failed\n", __FUNCTION__,__LINE__);
					return -1;
				}

				ret = union_cycle(graph, neighbor, vertex);
				if(ret != 0){
					printf("[%s][%d] union_cycle is failed\n", __FUNCTION__,__LINE__);
					return -1;
				}
			}
		}

		root = find(graph, vertex);

/* 4. remove impure members */
		for(i=0; i<vertex->degree; i++){
			if(vertex->neighbors[i] == root)
				continue;

			ret = hfind((htab *)graph->adj, &vertex->neighbors[i], sizeof(idx_t));
			if(ret == FALSE){
				printf("[%s][%d] hfind failed\n", __FUNCTION__,__LINE__);
				return -1;
			}
			neighbor = (struct vertex_t *)hstuff((htab *)graph->adj);

			idx_t temp =0;
			temp = find(graph, neighbor);
			if(temp == vertex->root){
				found = 0;
				ret = find_pure_community(graph, neighbor, &target, &found);
				if(ret != 0){
					printf("[%s][%d] find_pure_community failed\n", __FUNCTION__,__LINE__);
					return -1;
				}

				if((found == 0) || (target != temp)){
					neighbor->free = 1;
					ret = remove_root(graph, neighbor);
					if(ret != 0){
						printf("[%s][%d] remove_root failed\n", __FUNCTION__,__LINE__);
						return -1;
					}
					
					ret = remove_cycle(graph, neighbor);
					if(ret != 0){
						printf("[%s][%d] remove_cycle failed\n", __FUNCTION__,__LINE__);
						return -1;
					}
				}
			}
		}
		
/* 5. check the deeper members */
		for(i=0; i<vertex->degree; i++){
			ret = hfind((htab *)graph->adj, &vertex->neighbors[i], sizeof(idx_t));
			if(ret == FALSE){
				printf("[%s][%d] hfind failed\n", __FUNCTION__,__LINE__);
				return -1;
			}
			neighbor = (struct vertex_t *)hstuff((htab *)graph->adj);
			neighbor_root = find(graph, neighbor);

			if((neighbor->free != 1)&&(root == neighbor_root)){
				ret = update_neighbors(graph, neighbor);
				if(ret != 0){
					printf("[%s][%d] update_neighbors failed\n", __FUNCTION__,__LINE__);
					return -1;
				}
			}
		}
	}
 
	return 0;
}

void print_cycle(struct mep_graph_t *graph, struct vertex_t *vertex)
{
	int hret=0;
	struct vertex_t *v=0;

	v = vertex;
	while(v->cycle != vertex->key){
		printf("V%u(C%u)\n", v->key,find(graph, v));
		hret = hfind((htab *)graph->adj, &v->cycle, sizeof(idx_t));
		if(hret == FALSE){
			printf("[%s][%d] hfind failed\n", __FUNCTION__,__LINE__);
			return ;
		}
		v = hstuff((htab *)graph->adj);
	}
	printf("V%u(C%u)\n", v->key, find(graph, v));
}

int community_grow(struct mep_graph_t *graph)
{
	unsigned int i=0;
	int ret=0;
	int hret=0;
	struct vertex_t *vertex;

/* 1. choose the target vertex */
	for(i=0; i<graph->num_vertex; i++){
		hret = hfind((htab *)graph->adj, &graph->vertices[i], sizeof(idx_t));
		if(hret == FALSE){
			printf("[%s][%d] hfind failed\n", __FUNCTION__,__LINE__);
			ret = -1;
			goto out;
		}
		vertex = (struct vertex_t *)hstuff((htab *)graph->adj);

		if(vertex->free == 1){
			ret = update_vertex(graph, vertex);
			if(ret != 0){
				printf("[%s][%d] update_vertex failed\n", __FUNCTION__,__LINE__);
				ret = -1;
				goto out;
			}
		}
	}
	
out:
	return 0;
}

static int cleanup_sep_table(struct mep_graph_t *graph, struct vertex_t *vertex)
{
	idx_t *key=NULL;
	idx_t *value=NULL;

	if(vertex->sep != NULL){
		while(hcount((htab *)vertex->sep)){
			hfirst((htab *)vertex->sep);
			key = (idx_t *)hkey((htab *)vertex->sep);		
			value = (idx_t *)hstuff((htab *)vertex->sep);		
			if(key != NULL)
				free(key);
			if(value != NULL)
				free(value);

			hdel((htab *)vertex->sep);
		}
		hdestroy((htab *)vertex->sep);
	}
	vertex->sep = NULL;
	return 0;
}

int check_vertex_impurity(struct mep_graph_t *graph, struct vertex_t *vertex, idx_t *start_point_key)
{
	int ret=0;
	int hret=0;
	unsigned int found=0;
	idx_t old_root_key=0;
	idx_t new_root_key=0;
	struct vertex_t *new_root=NULL;

/* 1. check purity */
	hret = find_pure_community_new(graph, vertex, &new_root_key, &found);
	if(hret != 0){
		printf("[%s][%d] find_pure_community failed\n", __FUNCTION__,__LINE__);
		return -1;
	}
	
	if(!found){
		printf("[%s][%d] Something is wrong, there is no root!\n", __FUNCTION__,__LINE__);
		return -1;
	}
	old_root_key = find(graph, vertex);

/* 2. if impure, move the vertex to the pure community */
	if(old_root_key != new_root_key){/* impure! */
		if(vertex->key == *start_point_key){
			*start_point_key = vertex->cycle;
		}

/* 2.1. detach the impure vertex from its old root */
		if(vertex->key == vertex->root){
			ret = cleanup_sep_table(graph, vertex);
			if(ret != 0){
				printf("[%s][%d] cleanup_sep_table failed\n", __FUNCTION__,__LINE__);
				return -1;
			}
		}

		ret = remove_root(graph, vertex);
		if(ret != 0){
			printf("[%s][%d] remove_root failed\n", __FUNCTION__,__LINE__);
			return -1;
		}

		ret = remove_cycle(graph, vertex);
		if(ret != 0){
			printf("[%s][%d] remove_root failed\n", __FUNCTION__,__LINE__);
			return -1;
		}

/* 2.2. attache the impure vertex to its new root */
		hret = hfind((htab *)graph->adj, &new_root_key, sizeof(idx_t));
		if(hret == FALSE){
			printf("[%s][%d] hfind failed\n", __FUNCTION__,__LINE__);
			return -1;
		}
		new_root = hstuff((htab *)graph->adj);

		ret = union_remsp(graph, vertex, new_root);
		if(ret != 0){
			printf("[%s][%d] union_remsp failed\n", __FUNCTION__,__LINE__);
			return -1;
		}
		
		ret = union_cycle(graph, vertex, new_root);
		if(ret != 0){
			printf("[%s][%d] union_remsp failed\n", __FUNCTION__,__LINE__);
			return -1;
		}

		if(new_root->key != new_root->root){
			ret = cleanup_sep_table(graph, new_root);
			if(ret != 0){
				printf("[%s][%d] cleanup_sep_table failed\n", __FUNCTION__,__LINE__);
				return -1;
			}
		}
	}
#if 0
	else{
		vertex->comp_max = 0;
	}
#endif
	return 0;
}

int check_community_impurity(struct mep_graph_t *graph, struct vertex_t *root)
{
	int ret=0;
	int hret=0;
	struct vertex_t *vertex=NULL;
	idx_t next_key=0;
	idx_t start_point_key=0;
	
	if(root->key != root->root){
		hret = hfind((htab *)graph->comm, &root->key, sizeof(idx_t));
		if(hret == TRUE){
			hdel((htab *)graph->comm);
		}
		return 0;
	}

	vertex = root;
	start_point_key = root->key;

	while(vertex->cycle != start_point_key){
		next_key = vertex->cycle;
		ret = check_vertex_impurity(graph, vertex, &start_point_key);
		if(ret != 0){
			printf("[%s][%d] check_vertex_impurity failed\n", __FUNCTION__,__LINE__);
			return -1;
		}

		hret = hfind((htab *)graph->adj, &next_key, sizeof(idx_t));
		if(hret == FALSE){
			printf("[%s][%d] hfind failed\n", __FUNCTION__,__LINE__);
			return -1;
		}
		vertex = hstuff((htab *)graph->adj);
	}
	ret = check_vertex_impurity(graph, vertex, &start_point_key);
	if(ret != 0){
		printf("[%s][%d] check_vertex_impurity failed\n", __FUNCTION__,__LINE__);
		return -1;
	}

	return 0;
}

int community_merge(struct mep_graph_t *graph, idx_t *loops)
{
	unsigned int order=0;
	int ret=0;
	int hret=0;
	idx_t max_key=0;
	idx_t *key=NULL;
	float *value=NULL;
	float max_sep=0;
	unsigned int merge_count=0;
	struct vertex_t *Ci=NULL;
	struct vertex_t *Cj=NULL;
	struct vertex_t *temp=NULL;
	htab *impure_table=NULL;
	htab *pure_table=NULL;
	unsigned int count=0;

	if(loops == NULL){
		printf("[%s][%d] Invalid parameter\n", __FUNCTION__,__LINE__);
		return -1;
	}

	order = log((double)graph->num_comm);
	impure_table = hcreate(order+1);
	pure_table = hcreate(order+1);

/* 1. Main loop */
	while(1){
		count++;

		if(hcount(pure_table)>0){
			while(hcount(pure_table)){
				hfirst(pure_table);
				hdel(pure_table);
			}
			hdestroy(pure_table);
			pure_table = hcreate(order+1);
		}

/* Merging 1. Move the root vertices to the empty table first.
 * This is because it is possible for union function to change
 * the location of comm table in the union function.
 * Then the hfirst,hnext mechanism will miss some vertices. */ 
		if(hfirst((htab *)graph->comm)){
			do{
				Ci = hstuff((htab *)graph->comm);
				hret = hadd((htab *)pure_table, &Ci->key, sizeof(idx_t), (void *)Ci);
				if(hret == FALSE){
					printf("[%s][%d] hadd failed\n", __FUNCTION__,__LINE__);
					return -1;
				}
			}while(hnext((htab *)graph->comm));
		}

/* Merging 2. Traverse the root vertices and check if there is unbalanced community is.
 * If the compactness is smaller than the average separation, it is balanced. */
		if(hfirst((htab *)pure_table)){
			do{
				Ci = hstuff((htab *)pure_table);

				ret = update_community(graph, Ci, 0);
				if(ret != 0){
					printf("[%s][%d] update_community failed\n", __FUNCTION__,__LINE__);
					return -1;
				}

				if((Ci->compactness != 0) &&
					(Ci->compactness <= Ci->avg_sep)){/* unbalanced */
					max_sep = 0;
					if(hfirst((htab *)Ci->sep)){
						do{
							key = (idx_t *)hkey((htab *)Ci->sep);
							value = (float *)hstuff((htab *)Ci->sep);
							if(max_sep < *value){
								max_sep = *value;
								max_key = *key;
							}
						}while(hnext((htab *)Ci->sep));
					}
					else{
						printf("[%s][%d] C%u doesn't have sep table!\n", __FUNCTION__,__LINE__, Ci->key);
						return -1;
					}
					
					hret = hfind((htab *)graph->adj, &max_key, sizeof(idx_t));
					if(hret == FALSE){
						printf("[%s][%d] hfind failed\n", __FUNCTION__,__LINE__);
						return -1;
					}
					Cj = hstuff((htab *)graph->adj);

					ret = update_community(graph, Cj, 0);
					if(ret != 0){
						printf("[%s][%d] update_community failed\n", __FUNCTION__,__LINE__);
						return -1;
					}

					ret = union_remsp(graph, Ci, Cj);
					if(ret != 0){
						printf("[%s][%d] union_remsp failed\n", __FUNCTION__,__LINE__);
						ret = -1;
						goto out;
					}

					ret = union_cycle(graph, Ci, Cj);
					if(ret != 0){
						printf("[%s][%d] union_cycle failed\n", __FUNCTION__,__LINE__);
						ret = -1;
						goto out;
					}

					/* For readibility, merging community is Ci, merged community is Cj */
					if(Cj->key == Cj->root){
						temp = Cj;
						Cj = Ci;
						Ci = temp;
					}
					ret = cleanup_sep_table(graph, Cj);
					if(ret != 0){
						printf("[%s][%d] cleanup_sep_table failed\n", __FUNCTION__,__LINE__);
						ret = -1;
						goto out;
					}

					merge_count++;
				}
			}while(hnext((htab *)pure_table));
		}

		//printf("Merging %dth... merge_count: %u\n", count, merge_count);
/* Purity Check 1. Again, move all the roots into an empty table.
 * As explained above, union or delete function might effect the location of comm table. */
		if(hfirst((htab *)graph->comm)){
			do{
				Ci = hstuff((htab *)graph->comm);
				hret = hfind(impure_table, &Ci->key, sizeof(idx_t));
				if(hret == FALSE){
					hret = hadd(impure_table, &Ci->key, sizeof(idx_t), (void *)Ci);
					if(hret == FALSE){
						printf("[%s][%d] hadd failed\n", __FUNCTION__,__LINE__);
						return -1;
					}
				}
			}while(hnext((htab *)graph->comm));
		}

/* Purity Check 2. Traverse every vertex and check if they are pure. */
		if(hfirst(impure_table)){
			do{
				Ci = hstuff(impure_table);

				ret = check_community_impurity(graph, Ci);
				if(ret != 0){
					printf("[%s][%d] check_community_impurity failend\n", __FUNCTION__,__LINE__);
					ret = -1;
					goto out;
				}
			}while(hnext(impure_table));
		}

		while(hcount(impure_table)){
			hfirst(impure_table);
			hdel(impure_table);
		}

		if(merge_count == 0)
			break;
		else
			merge_count = 0;
	}
	hdestroy(impure_table);

out:
	*loops = count;
	return ret;
}

int calculate_purity_values(struct mep_graph_t *graph)
{
	int ret=0;
	unsigned int num_pures=0;
	unsigned int num_members=0;
	idx_t pure_community=0;
	struct vertex_t *root=NULL;
	struct vertex_t *vertex=NULL;

	if(hfirst((htab *)graph->comm)){
		do{
			root = hstuff((htab *)graph->comm);

			vertex = root;
			num_pures = 0;
			num_members = 0;
			while(vertex->cycle != root->key){
				num_members++;

				pure_community = find(graph, vertex);
				if(pure_community == root->key)
					num_pures++;

				ret = hfind((htab *)graph->adj, &vertex->cycle, sizeof(idx_t));
				if(ret == FALSE){
					printf("[%s][%d] hfind failed\n", __FUNCTION__,__LINE__);
					return -1;
				}
				vertex = hstuff((htab *)graph->adj);
			}
			num_members++;

			pure_community = find(graph, vertex);
			if(pure_community == root->key)
				num_pures++;
			
			if(num_members == 0){
				printf("[%s][%d] num_member is zero... something is wrong..\n", __FUNCTION__,__LINE__);
				return -1;
			}
			root->purity = (float)num_pures/(float)num_members;
		}while(hnext((htab *)graph->comm));
	}

	return 0;
}

int printout(char *filebasename, struct mep_graph_t *graph, idx_t loops)
{
	unsigned int i=0;
	int ret=0;
	int hret=0;
	FILE *fd=NULL;
	struct vertex_t *vertex=NULL;
	char filename[MAX_STRING_LENGTH];
	char *output=NULL;
	char *previous=NULL;

	output = strtok(filebasename, "/");
	while(output != NULL){
		previous = output;
		output = strtok(NULL, "/");
	}
	
	sprintf(filename, "%s_comm1.txt", previous);

	fd = fopen(filename, "w+");
	if(fd == NULL){
		printf("[%s][%d] fopen failed\n", __FUNCTION__,__LINE__);
		ret = -1;
		goto out;
	}

	for(i=0; i<graph->num_vertex; i++){
		hret = hfind((htab *)graph->adj, &graph->vertices[i], sizeof(idx_t));
		if(hret == FALSE){
			printf("[%s][%d] hfind failed\n", __FUNCTION__,__LINE__);
			fclose(fd);
			ret = -1;
			goto out;
		}
		
		vertex = hstuff((htab *)graph->adj);
		fprintf(fd, "%u	%u\n", vertex->key, find(graph, vertex));
	}
	
	fflush(fd);
	fclose(fd);

	fd = fopen("merging_counts.txt", "w+");
	if(fd == NULL){
		printf("[%s][%d] fopen failed\n", __FUNCTION__,__LINE__);
		ret = -1;
		goto out;
	}

	fprintf(fd, "%u\n", loops);
	fflush(fd);
	fclose(fd);
	fd = NULL;

out:
	if(fd != NULL)
		fclose(fd);
	return ret;
}
