/*
 *
 * This file contains routines related to mep algorithm
 *
 * Diana Palsetia
 * 
 * Copyright (C) 2016, Northwestern University
 * See COPYRIGHT notice in top-level directory.
 *
 */
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <malloc.h>
#include <string.h>
#include <sys/time.h>
#include <math.h>
#include "pmep.h"

double wtime(void) /* to be used when MPI_Wtime is not available */
{
    double          now_time;
    struct timeval  etstart;
    struct timezone tzp;

    if (gettimeofday(&etstart, &tzp) == -1)
        perror("Error: calling gettimeofday() not successful.\n");

    now_time = ((double)etstart.tv_sec) +              /* in seconds */
               ((double)etstart.tv_usec) / 1000000.0;  /* in microseconds */
    return now_time;
}

static int compare(const void *a, const void *b)
{
    idx_t ret = *(idx_t*)a - *(idx_t*)b;
    return (int)ret;
}

int create_graph(idx_t               num_edges,
                 idx_t              *edges,      /* [num_edges] */
	         idx_t               direction,
                 struct mep_graph_t *pgraph)
{
	int i=0, j=0, k=0;
	int ret=0;
	int hret=0;
	idx_t num_vertex=0;
	idx_t *vertices=NULL;
	idx_t previous=0;
	double order=0;

	pgraph->num_edge = num_edges/2;

/* 1. find unique vertex list */
	vertices = (idx_t *)malloc(sizeof(idx_t)*num_edges);
	if(vertices == NULL){
		printf("[%s][%d] out of memory\n", __FUNCTION__,__LINE__);
		return -1;
	}
	memcpy(vertices, edges, sizeof(idx_t)*num_edges);

    /* sort vertex list in an increasing order */
#if USE_QSORT
	qsort(vertices, num_edges, sizeof(int), compare);
#else
	ret = countingSort(vertices, NULL, num_edges, 0);
	if(ret != 0){
		printf("[%s][%d] countingSort failed\n", __FUNCTION__,__LINE__);
		return -1;
	}
#endif

	/* remove deplicated vertices */
	previous = vertices[0];
	num_vertex = 1;
	for (i=1; i<num_edges; i++) {
		if (vertices[i] != previous) {
			vertices[num_vertex++] = vertices[i];
			previous = vertices[i];
		}
	}
	/* num_vertex is now the number of unique vertices assigned to this process */
	if (num_vertex != num_edges)
	    vertices = (idx_t*) realloc(vertices, sizeof(idx_t)*num_vertex);

	pgraph->vertices = vertices;

/* 2. initialize adj table */
	pgraph->num_vertex = num_vertex;
	order = log((double)num_vertex);
	pgraph->adj = hcreate((idx_t)order+1);
	pgraph->comm = hcreate((idx_t)order+1);
	pgraph->num_comm=0;

/* 3. build vertex table */
/* 3.1. allocate vertex table */
	pgraph->vtable = (struct vertex_t *)malloc(sizeof(struct vertex_t)*num_vertex);
	if(pgraph->vtable == NULL){
		printf("[%s][%d] out of memory\n", __FUNCTION__,__LINE__);
		return -1;
	}

/* 3.2. insert every vertex into the hash table */
	for(i=0; i<num_vertex; i++){
		pgraph->vtable[i].key = vertices[i];
		pgraph->vtable[i].degree = 0;
		pgraph->vtable[i].temp_num = 0;
		pgraph->vtable[i].sep = NULL;
		hret = hadd((htab *)pgraph->adj, &(pgraph->vtable[i].key), sizeof(idx_t), (void *)&(pgraph->vtable[i]));
		if(hret == FALSE){
			printf("[%s][%d] hadd failed\n", __FUNCTION__,__LINE__);
			free(pgraph->vtable);
			return -1;
		}
	}

/* 3.3. get the degree of each vertex */
	if(direction == 1)
		direction = 2;
	else
		direction = 1;

	for(i=0; i<num_edges; i+=2){
		for(j=0; j<direction; j++){
			hret = hfind((htab *)pgraph->adj, (ub1 *)&edges[i+j], sizeof(idx_t));
			if(hret == FALSE){
				printf("[%s][%d] hfind failed\n", __FUNCTION__,__LINE__);
				free(pgraph->vtable);
				return -1;
			}
			struct vertex_t *vertex = (struct vertex_t *)hstuff((htab *)pgraph->adj);
			vertex->degree++;
		}
	}

/* 3.4. make the neighbor lists */
	for(i=0; i<num_vertex; i++){
		hret = hfind((htab *)pgraph->adj, (ub1*)(vertices+i), sizeof(idx_t));
		if(hret == FALSE){
			printf("[%s][%d] hfind failed\n", __FUNCTION__,__LINE__);
			free(pgraph->vtable);
			return -1;
		}
		struct vertex_t *vertex = (struct vertex_t *)hstuff((htab *)pgraph->adj);
		vertex->neighbors = (idx_t *)malloc(sizeof(idx_t)* vertex->degree);
		if(vertex->neighbors == NULL){
			printf("[%s][%d] out of memory\n", __FUNCTION__,__LINE__);
			free(pgraph->vtable);
			return -1;
		}
	}

/* 3.4. building an adjacency list */
	for(i=0; i<num_edges; i+=2){
		for(j=0; j < direction; j++){
			hret = hfind((htab *)pgraph->adj, (ub1 *)&edges[i+j], sizeof(idx_t));
			if(hret == FALSE){
				printf("[%s][%d] hfind failed\n", __FUNCTION__,__LINE__);
				free(pgraph->vtable);
				return -1;
			}
			struct vertex_t *vertex = (struct vertex_t *)hstuff((htab *)pgraph->adj);
			k = (j == 0)?1:0;
			vertex->neighbors[vertex->temp_num++] = edges[i+k];
		}
	}
	
	return 0;
}

int build_free_list(struct mep_graph_t *graph)
{
	int ret=0;
	int hret=0;
	htab *table=NULL;
	int i=0;
	idx_t *degrees=NULL;
	struct vertex_t *vertex=NULL;

/* sort it by degree */
	degrees = (idx_t *)malloc(sizeof(idx_t)*graph->num_vertex);
	if(degrees == NULL){
		printf("[%s][%d] out of memory\n", __FUNCTION__,__LINE__);
		return -1;
	}

	table = (htab *)graph->adj;
	for(i=0; i<graph->num_vertex; i++){
		hret = hfind(table, (ub1 *)&graph->vertices[i], sizeof(idx_t));
		if(hret == FALSE){
			printf("[%s][%d] hfind failed\n", __FUNCTION__,__LINE__);
			return -1;
		}
		vertex = (struct vertex_t *)hstuff(table);
		degrees[i] = vertex->degree;
		vertex->free = 1;
		vertex->root = vertex->key;
		vertex->cycle = vertex->key;
	}
	
        /* sort vertixec by their degrees */
	ret = countingSort(degrees, graph->vertices, graph->num_vertex, 1);
	if(ret != 0){
		printf("[%s][%d] countingSort failed\n", __FUNCTION__,__LINE__);
		return -1;
	}
	free(degrees);
	return 0;
}

void cleanup_table(struct htab *table)
{
	idx_t *key=NULL;
	idx_t *data=NULL;

	if(hfirst(table)){
		do{
			key = NULL;
			data = NULL;
			key = (idx_t *)hkey(table);
			data = (idx_t *)hstuff(table);
			if(key != NULL)
				free(key);
			if(data != NULL)
				free(data);
		}while(hnext(table));
	}
	hdestroy(table);
}

void cleanup_vertex_table(struct mep_graph_t *graph)
{
	idx_t *key = NULL;
	struct vertex_t *vertex=NULL;

	if(graph != NULL){
/* 1. get every vertex */
		if(hfirst((htab *)graph->adj)){
			do{
				vertex = hstuff((htab *)graph->adj);
/* 2. get every sep table from each vertex */
				if(vertex->key == vertex->root){
					if(vertex->sep == NULL){
						//printf("[%s][%d] V%u is root but sep is null..\n", __FUNCTION__,__LINE__, vertex->key);
					}
					else{
						if(hfirst((htab *)vertex->sep)){
							do{
								key = (idx_t *)hkey((htab *)vertex->sep);
								if(key != NULL){
									free(key);
									key =NULL;
								}
							}while(hnext((htab *)vertex->sep));
						}
						hdestroy((htab *)vertex->sep);
					}
				}
			}while(hnext((htab *)graph->adj));
		}
		hdestroy((htab *)graph->adj);
		hdestroy((htab *)graph->comm);
	}
}

void delete_graph(struct mep_graph_t *graph)
{
	cleanup_vertex_table(graph);

	if(graph->vtable != NULL){
		free(graph->vtable);
		graph->vtable = NULL;
	}
}

int c_serial_version_main(int argc, char **argv)
{
	int ret=0;
	int opt=0;
	idx_t loops=0;
	char *filename=NULL;
	idx_t filetype=2;
	struct mep_graph_t graph;
	double t0, t1, t2, total=0;

/* 1. parameter handling */
	while((opt = getopt(argc, argv, "i:t:"))!= -1){
		switch(opt)
		{
			case 'i':
				filename = optarg;
				printf("filename: %s\n", filename);
				break;
			case 't':
				filetype = atoi(optarg);
				printf("filetype: %s\n", (filetype == 0)?"binary":"txt");
				break;
			default:
				printf("Invalid option!\n");
				return -1;
		}
	}

	if((filename == NULL) || (filetype > 1)){
		printf("usage: ./test -i [filename] -t [filetype(0: binary, 1: txt)]\n");
		return 0;
	}

	t0 = t1 = wtime();
/* 2. read the input data */
        idx_t  direction;
        idx_t  num_edges;
        idx_t *edges;
        ret = init_data_with_file(filename, filetype, &direction, &num_edges, &edges);
	if(ret != 0){
		printf("[%s][%d] %s init_graph failed\n", __FUNCTION__,__LINE__,filename);
		return -1;
	}
	t2 = wtime();
	total += (t2-t1);
	printf("File Reading: %f sec\n", t2-t1);

	t1 = wtime();
/* 3. build adjacency list & free list  */
        /* TODO: check if the 1st argument should use num_edges*2 or num_edges */
        ret = create_graph(num_edges*2, edges, direction, &graph);
	if(ret != 0){
		printf("[%s][%d] build_graph failed\n", __FUNCTION__,__LINE__);
		return -1;
	}
        free(edges);
	t2 = wtime();
	total += (t2-t1);
	printf("Creating Graph: %f sec\n", t2-t1);

	t1 = wtime();
	ret = build_free_list(&graph);
	if(ret != 0){
		printf("[%s][%d] build_free_list failed\n", __FUNCTION__,__LINE__);
		return -1;
	}
	
	ret = community_grow(&graph);
	if(ret != 0){
		printf("[%s][%d] community_grow failed\n", __FUNCTION__,__LINE__);
		return -1;
	}
	t2 = wtime();
	total += (t2-t1);
	printf("Community Growing Phase: %f sec\n", t2-t1);

	t1 = wtime();
	ret = community_merge(&graph, &loops);
	if(ret != 0){
		printf("[%s][%d] community_merge failed\n", __FUNCTION__,__LINE__);
		return -1;
	}
	t2 = wtime();
	total += (t2-t1);
	printf("Community Merging Phase: %f sec\n", t2-t1);
	printf("Total: %f sec\n", total);

	t1 = wtime();
	printout(filename, &graph, loops);
	t2 = wtime();
	total += (t2-t1);

/* 4. free graph object */
	t1 = wtime();
	delete_graph(&graph);
	t2 = wtime();
	total += (t2-t1);
	printf("Total with Writing: %f sec\n", total);

	return 0;
}
