/*
 *
 * This file contains routines related to parallel mep algorithm
 *
 * Diana Palsetia
 * 
 * Copyright (C) 2016, Northwestern University
 * See COPYRIGHT notice in top-level directory.
 *
 */

#include "pmep.h"

#define    MAXLINE    64*1024*1024

/*************************************************************************
* This function reads the binary CSR matrix and set the following variables.
  gnvtxs: total number of vertices
  gnedges: total number of edges
  nvtxs: number of vertices assigned to this process
  nedges: number of edges (adjacent vertices) of vertices assigned to
          this process
  vtxdist[npes+1]: vertex distributed offsets, starting vertex ID of
          assigned vertices
  xadj[nvtxs+1]: exclusive scan of number of edges per vertex
  adjncy[nvtxs]: list of adjacent vertices assigned to this process
**************************************************************************/
void ParallelReadGraph2(char      *fname_bin,
                        char      *fname_off,
                        idx_t     *gnvtxs,
                        idx_t     *gnedges,
                        idx_t     *nvtxs,
                        idx_t     *nedges,
                        idx_t    **vtxdist,
                        idx_t    **xadj,
                        idx_t    **adjncy,
                        MPI_Comm   comm)
{
    int i, npes, mype, err;
    idx_t offset, header[2];
    MPI_File fhOff, fhBin;
    MPI_Status st;

    MPI_Comm_size(comm, &npes);
    MPI_Comm_rank(comm, &mype);
    *vtxdist = (idx_t*) calloc(npes+1, sizeof(idx_t));

    /* open 2 files: vertex offset file and adjacency file */
    err = MPI_File_open(comm, fname_bin, MPI_MODE_RDONLY, MPI_INFO_NULL, &fhBin);
    if (err != MPI_SUCCESS) {
        int errorStringLen;
        char errorString[MPI_MAX_ERROR_STRING];
        MPI_Error_string(err, errorString, &errorStringLen);
        printf("Error: MPI_File_open() %s (%s)\n",fname_bin, errorString);
        MPI_Abort(comm, -1);
    }
    err = MPI_File_open(comm, fname_off, MPI_MODE_RDONLY, MPI_INFO_NULL, &fhOff);
    if (err != MPI_SUCCESS) {
        int errorStringLen;
        char errorString[MPI_MAX_ERROR_STRING];
        MPI_Error_string(err, errorString, &errorStringLen);
        printf("Error: MPI_File_open() %s (%s)\n",fname_off, errorString);
        MPI_Abort(comm, -1);
    }
    if (mype == 0) { /* read header: total number of vertices and edges */
        err = MPI_File_read(fhBin, header, 2, MPI_INT, &st);
        if (err != MPI_SUCCESS) {
            int errorStringLen;
            char errorString[MPI_MAX_ERROR_STRING];
            MPI_Error_string(err, errorString, &errorStringLen);
            printf("Error: MPI_File_read() %s (%s)\n",fname_bin, errorString);
            MPI_Abort(comm, -1);
        }
    }
    MPI_Bcast(header, 2, MPI_INT, 0, comm);
    *gnvtxs  = header[0];
    *gnedges = header[1] * 2;

	
    /* calculate (*vtxdist)[] */
    int l,k;
    (*vtxdist)[0] = 0;
	for (i=0,k=*gnvtxs; i<npes; i++) {
      l = k/(npes-i);
      (*vtxdist)[i+1] = (*vtxdist)[i]+l;
	  k -= l;
    }
	
    /* process mype is assigned vertices from (*vtxdist)[mype] to
     * (*vtxdist)[mype+1] */
    assert(*gnvtxs == (*vtxdist)[npes]);
    *nvtxs = (*vtxdist)[mype+1] - (*vtxdist)[mype];
    *xadj  = (idx_t*) malloc((*nvtxs+1) * sizeof(idx_t));

    /* *nvtxs is the number of vertices assigned to this process
     * (*xadj)[i] is the number of adjancet neighbors of vertex i
     */

    idx_t *off_v = (idx_t*) malloc((*nvtxs+1) * sizeof(idx_t));
    offset = (*vtxdist)[mype] * sizeof(idx_t); /* offset to offset file */
    err = MPI_File_read_at_all(fhOff, offset, off_v, *nvtxs+1, MPI_INT, &st);
    if (err != MPI_SUCCESS) {
        int errorStringLen;
        char errorString[MPI_MAX_ERROR_STRING];
        MPI_Error_string(err, errorString, &errorStringLen);
        printf("Error: MPI_File_read_at_all() %s (%s)\n",fname_off, errorString);
        MPI_Abort(comm, -1);
    }
    for (i=0; i<*nvtxs; i++)
        (*xadj)[i] = (off_v[i+1] - off_v[i]) / sizeof(idx_t);

    /* execlusive scan on (*xadj)[] */
    for (i=1; i<*nvtxs; i++) (*xadj)[i] += (*xadj)[i-1];
    for (i=*nvtxs; i>0; i--) (*xadj)[i]  = (*xadj)[i-1];
    (*xadj)[0] = 0;
    /* now (*xadj)[*nvtxs] is the number of all adjancy neighbors of all vertice */

    *nedges = (*xadj)[*nvtxs];
    *adjncy = (idx_t*) malloc(*nedges * sizeof(idx_t));
    /* the adjancy vertices are from file offset off_v[0] to off_v[*nvtxs+1]
     * in file_bin*/
    assert(*nedges == (off_v[*nvtxs] - off_v[0]) / sizeof(idx_t));
    offset = off_v[0];
    err = MPI_File_read_at_all(fhBin, offset, *adjncy, *nedges, MPI_INT, &st);
    if (err != MPI_SUCCESS) {
        int errorStringLen;
        char errorString[MPI_MAX_ERROR_STRING];
        MPI_Error_string(err, errorString, &errorStringLen);
        printf("Error: MPI_File_read_at_all() %s (%s)\n",fname_bin, errorString);
        MPI_Abort(comm, -1);
    }
    /* make the vertex IDs start at 0 */
    for (i=0; i<*nedges; i++) (*adjncy)[i]--;
    free(off_v);

    /* close files */
    err = MPI_File_close(&fhOff);
    if (err != MPI_SUCCESS) {
        int errorStringLen;
        char errorString[MPI_MAX_ERROR_STRING];
        MPI_Error_string(err, errorString, &errorStringLen);
        printf("Error: MPI_File_close() %s (%s)\n",fname_off, errorString);
        MPI_Abort(comm, -1);
    }
    err = MPI_File_close(&fhBin);
    if (err != MPI_SUCCESS) {
        int errorStringLen;
        char errorString[MPI_MAX_ERROR_STRING];
        MPI_Error_string(err, errorString, &errorStringLen);
        printf("Error: MPI_File_close() %s (%s)\n",fname_bin, errorString);
        MPI_Abort(comm, -1);
    }
}


idx_t get_pid(idx_t * vtxdist,idx_t v,idx_t npes)
{
    //*************************************************************************
    // This function return process id that will contain part id of vertex v
    //**************************************************************************
    idx_t pid,div,rem;
    pid = -1;
    
    div = vtxdist[npes]/npes;
    rem =  vtxdist[npes] % npes;
    if (v < div*(npes-rem))
        pid = v/div;
    else
        pid = (v+npes-rem)/(div+1);
    return pid;
}
    



void exchange_partitions(graph_t * graph, idx_t * part, MPI_Comm comm, idx_t* recv_part)
{    
    /*************************************************************************
    * This function exchanges partition data
    **************************************************************************/
    int mype, npes;
    idx_t p,offset,msize;
    
    MPI_Comm_size(comm, &npes);
    MPI_Comm_rank(comm, &mype);
    
    MPI_Request * requests;
    MPI_Status  * status;

    
    requests = (MPI_Request*)malloc((npes * 2)*sizeof(MPI_Request));
    status   =  (MPI_Status*)malloc((npes * 2)*sizeof(MPI_Status));
    
    for (p=0; p<npes; p++)
    {
        offset = graph->vtxdist[p];
        msize = graph->vtxdist[p+1] - graph->vtxdist[p];
        MPI_Irecv((void *)(recv_part+offset),(idx_t)msize,IDX_T,p,1,MPI_COMM_WORLD,&(requests[p]));
    }
    for (p=0; p<npes; p++)
    {
        MPI_Isend((void *)part,(idx_t)graph->nvtxs,IDX_T,p,1,MPI_COMM_WORLD,&(requests[p+npes]));
    }
    
    int err = MPI_Waitall(npes*2, requests, status);
    if (err != MPI_SUCCESS) {
        char err_string[MPI_MAX_ERROR_STRING+1];
        int  i, err_len;
        MPI_Error_string(err, err_string, &err_len);
        printf("Error: MPI_Waitall() at %s line %d (%s)\n",__func__,__LINE__,err_string);
        for (i=0; i<npes*2; i++)
            if (status[i].MPI_ERROR != MPI_SUCCESS) {
                MPI_Error_string(err, err_string, &err_len);
                printf("Error: MPI_Waitall() at %s line %d (status[%d] %s)\n",__func__,__LINE__,i,err_string);
            }
    }
    
    free(requests);
    free(status);
    
}

void create_overlaps(graph_t * graph, idx_t nparts, idx_t * recv_part , idx_t ** pedges, idx_t * count, MPI_Comm comm)
{
    /*************************************************************************
    * This function creates overlaps from partition data & original edge list
    * Each process will have partial edge list for every other process
    **************************************************************************/
    int mype, npes;
    idx_t i,j,deg,offset;
    MPI_Comm_size(comm, &npes);
    MPI_Comm_rank(comm, &mype);
    
    idx_t num_vertices = graph->nvtxs;
    idx_t left, right,u,v,pid;
    
    idx_t * pcount = (idx_t *)calloc(nparts,sizeof(idx_t));
    idx_t ** parts = (idx_t**)malloc(nparts*sizeof(idx_t*));
    for(i = 0; i < nparts;i++){
        parts[i] = (idx_t*)malloc(graph->nedges*sizeof(idx_t));
    }
    
    for( i = 0; i < num_vertices;i++)
    {
        u = i + graph->vtxdist[mype];
        deg = graph->xadj[i+1]-graph->xadj[i]; 
        offset = graph->xadj[i];
        pid = -1;
        for(j = offset; j < offset+deg;j++)
        {
            v = graph->adjncy[j];
                
            if (u > v)
                continue;
            
            left = recv_part[u];
            right = recv_part[v];
            if (left == right)
            {
                parts[left][pcount[left]] = u;
                pcount[left]++;
                parts[left][pcount[left]] = v;
                pcount[left]++;
                pid = -1;
            }
            else
            {
                if (right > left)
                    pid = (right*(right+1))/2 - left - 1;
                else
                    pid = (left*(left+1))/2 - right - 1;
                pedges[pid][count[pid]] = u;
                count[pid]++;
                pedges[pid][count[pid]] = v;
                count[pid]++;
                
            }
            
        }
    }    
    
    
    for(i=0;i < npes;i++)
    {
        
        double a = 1.0  + sqrt(8*i + 1.0);
        right = (idx_t)floor(a*0.5);
        left = right + ((right*(right-1))/2) - i - 1;
    }
}

void exchange_partitions2_compute(idx_t     *dp_count,
                                  idx_t    **dpid,
                                  graph_t   *graph,
                                  idx_t     *part,
								  htab 		*dups,
                                  MPI_Comm   comm)
{
    /*************************************************************************
    * This function exchanges partition data - Distributed
    **************************************************************************/
    int mype, npes;
    idx_t i,j,u,v,pid,deg,n_size,offset;//recv_offset,send_offset;
   
    idx_t *key = NULL;
    unsigned int keyl = sizeof(idx_t);
    idx_t *data = NULL;
    
    int ret=0;
    idx_t num_vertices = graph->nvtxs;
    idx_t one=1;
    
    MPI_Comm_size(comm, &npes);
    MPI_Comm_rank(comm, &mype);
    
    
    if(dups == NULL)
	{
		printf("Could not allocate htable\n");
		MPI_Abort(MPI_COMM_WORLD, -1);
		 
	}
    
    for( i = 0; i < num_vertices;i++)
    {
        u = i + graph->vtxdist[mype];
        dpid[mype][dp_count[mype]] = u;
        dp_count[mype]++;

       
	   //Go through u's neighbors and figure out which process we need to communicate with
        deg = graph->xadj[i+1]-graph->xadj[i]; 
        offset = graph->xadj[i];
        n_size = offset + deg;
        for(j = offset; j < n_size;j++)
        {
            v = graph->adjncy[j];
           
            if (u < v)
            {
                 
				ret = hfind(dups,(ub1 *)&v,keyl); //to keep track of unique v 
				if(ret == FALSE)
                {
                    
					
                    pid = get_pid(graph->vtxdist,v,npes);
					if(pid >= npes)
					{
						printf("Error with pid mapping\n");
						MPI_Abort(comm, -1);
					}
					
                    if (pid != mype)
                    {
                        dpid[pid][dp_count[pid]] = v;
                        dp_count[pid]++;
                    }
					
					key  = (idx_t *)malloc(sizeof(idx_t)*1);
					data = (idx_t *)malloc(sizeof(idx_t)*1);
					*key = v;
					data[0] = 1;
                    hadd(dups,(ub1 *)key,keyl,data);
                }
            }
        }
    }
}

htab * exchange_partitions2_comm(idx_t * dp_count, idx_t ** dpid,graph_t * graph, idx_t * part, MPI_Comm comm)
{    
    /*************************************************************************
    * This function exchanges partition data - Distributed
    **************************************************************************/
    int mype, npes;
    idx_t i,j,v,p,recv_offset,send_offset;
    idx_t * recv_part = NULL;
    //htab *dups=NULL,
    htab * new_part=NULL;
    idx_t *key = NULL;
    unsigned int keyl = sizeof(idx_t);
    idx_t *data = NULL;
    idx_t hsize;
    //int ret=0;
    

    MPI_Comm_size(comm, &npes);
    MPI_Comm_rank(comm, &mype);
    
    MPI_Request *requests;
    MPI_Status  * status;

    
    //Communicate number of vertices to be received from all other processes
    idx_t * s_count = (idx_t *)calloc(npes,sizeof(idx_t));
    idx_t * r_count = (idx_t *)calloc(npes,sizeof(idx_t));

    for(p=0;p < npes;p++)
    {
        
        s_count[p] = dp_count[p];
    }
    
    MPI_Alltoall((void *)s_count,1,IDX_T,(void *)r_count, 1, IDX_T,MPI_COMM_WORLD);
    
    //Communicate the vertices
    idx_t total_v = 0;
    for(p = 0;p<npes;p++)
    {
        if(mype != p)
            total_v += r_count[p];
    }
    idx_t * recv_v = (idx_t *)malloc(total_v*sizeof(idx_t));
    
    requests = (MPI_Request*)malloc((npes * 2)*sizeof(MPI_Request));
    status   = (MPI_Status*)malloc((npes * 2)*sizeof(MPI_Status));
    
    recv_offset = 0;
    
    for(p=0;p < npes;p++)
    {
        requests[p]=MPI_REQUEST_NULL;
        if((mype != p) && (r_count[p] > 0))
        {
            MPI_Irecv((void *)(recv_v+recv_offset),r_count[p],IDX_T,p,1,MPI_COMM_WORLD,&requests[p]);
            recv_offset += r_count[p];
        }
    }

    for(p=0;p < npes;p++)
    {
        requests[p+npes]=MPI_REQUEST_NULL;
        if((mype != p) && (s_count[p] > 0)){
            MPI_Isend((void *)dpid[p],s_count[p],IDX_T,p,1,MPI_COMM_WORLD,&requests[p+npes]);
        }
    }
    
    int err = MPI_Waitall(npes*2, requests, status);
    if (err != MPI_SUCCESS) {
        char err_string[MPI_MAX_ERROR_STRING+1];
        int  i, err_len;
        MPI_Error_string(err, err_string, &err_len);
        printf("Error: MPI_Waitall() at %s line %d (%s)\n",__func__,__LINE__,err_string);
        for (i=0; i<npes*2; i++)
            if (status[i].MPI_ERROR != MPI_SUCCESS) {
                MPI_Error_string(err, err_string, &err_len);
                printf("Error: MPI_Waitall() at %s line %d (status[%d] %s)\n",__func__,__LINE__,i,err_string);
            }
    }
    
    
    
    
    //Communicate the partition_ids
    idx_t * send_part = (idx_t*)malloc(total_v*sizeof(idx_t));

    for(i=0;i < total_v;i++)
    {
        v = recv_v[i] - graph->vtxdist[mype]; 
        send_part[i] = part[v];
        
    }
      
     
    for(p = 0;p<npes;p++)
    {
        if(mype != p)
            total_v += s_count[p];
    }
    recv_part = (idx_t *)malloc(total_v*sizeof(idx_t));
    
    recv_offset = 0;
    send_offset = 0;
    
    for(p=0;p < npes;p++)
    {
        requests[p]=MPI_REQUEST_NULL;
        if((mype != p) && (s_count[p] > 0))
        {
            MPI_Irecv((void *)(recv_part+recv_offset),s_count[p],IDX_T,p,1,MPI_COMM_WORLD,&(requests[p]));
            recv_offset += s_count[p];
        }

    }

    for(p=0;p < npes;p++)
    {
        requests[p+npes]=MPI_REQUEST_NULL;
        if((mype != p) && (r_count[p] > 0))
        {
            MPI_Isend((void *)(send_part+send_offset),r_count[p],IDX_T,p,1,MPI_COMM_WORLD,&(requests[p+npes]));
            send_offset += r_count[p];
        }
    }
    
    err = MPI_Waitall(npes*2, requests, status);
    if (err != MPI_SUCCESS) {
        char err_string[MPI_MAX_ERROR_STRING+1];
        int  i, err_len;
        MPI_Error_string(err, err_string, &err_len);
        printf("Error: MPI_Waitall() at %s line %d (%s)\n",__func__,__LINE__,err_string);
        for (i=0; i<npes*2; i++)
            if (status[i].MPI_ERROR != MPI_SUCCESS) {
                MPI_Error_string(err, err_string, &err_len);
                printf("Error: MPI_Waitall() at %s line %d (status[%d] %s)\n",__func__,__LINE__,i,err_string);
            }
    }
    
    total_v += dp_count[mype];
    hsize = (idx_t)(log(total_v)/log(2));
    new_part = hcreate(hsize);
    
    j = 0;
    for(p = 0;p < npes;p++)
    {
        if (p == mype) 
        {
            for(i=0;i< dp_count[p];i++){ 
                key = (idx_t *)malloc(sizeof(idx_t)*1);
                data = (idx_t *)malloc(sizeof(idx_t)*1);
                *key = dpid[p][i];
                data[0] = part[dpid[p][i] - graph->vtxdist[mype]];
                hadd(new_part, (ub1 *)key, keyl, data);
            }
        }
        else{
            for(i=0;i< dp_count[p];i++){
                key = (idx_t *)malloc(sizeof(idx_t)*1);
                data = (idx_t *)malloc(sizeof(idx_t)*1);
                *key = dpid[p][i];
                data[0] = recv_part[j];
                hadd(new_part, (ub1 *)key, keyl, data);
                j++;
            }
        }
    }
    
    free(s_count);
    free(r_count);
    free(recv_v);
    free(recv_part);
    free(send_part);
    free(requests);
    free(status);
    
    
    return new_part;
}


idx_t** create_overlaps2(graph_t * graph, idx_t nparts, htab * new_part,idx_t ** bedges, idx_t * bcount, idx_t ** wedges, idx_t * wcount, idx_t * pcount, MPI_Comm comm)
{
    /*************************************************************************
    * This function creates overlaps from partition data & original edge list
    * Each process will have partial edge list for every other process
    **************************************************************************/
    int mype, npes;
    idx_t i,j,deg,offset;
    MPI_Comm_size(comm, &npes);
    MPI_Comm_rank(comm, &mype);
    
    idx_t num_vertices = graph->nvtxs;
    idx_t left, right,u,v,pid;
    
    unsigned int keyl=sizeof(idx_t);
    idx_t *newdata=NULL;
    int ret = 0;
    

    idx_t count = 0;
    for( i = 0; i < num_vertices;i++)
    {
        u = i + graph->vtxdist[mype];
        deg = graph->xadj[i+1]-graph->xadj[i]; 
        offset = graph->xadj[i];
        pid = -1;
        
        for(j = offset; j < offset+deg;j++)
        {
            v = graph->adjncy[j];
                
            if (u < v)
            {
                count++;
                ret = hfind(new_part, &u, keyl);
                if(ret == FALSE)
                {
                    printf("[%s][%d] there is no item which has a key like %d\n", __FUNCTION__,__LINE__, u);
                    MPI_Abort(MPI_COMM_WORLD, 1);
                }
                else
                {
                    newdata = hstuff(new_part);
                    left = newdata[0];
                }
                ret = hfind(new_part, &v, keyl);
                if(ret == FALSE){
                    printf("[%s][%d] there is no item which has a key like %d\n", __FUNCTION__,__LINE__, v);
                    MPI_Abort(MPI_COMM_WORLD, 1);
                }
                else
                {
                    newdata = hstuff(new_part);
                    right = newdata[0];
                }
                if (left == right)
                { 
                    pid = -1;
                    wedges[left][wcount[left]] = u+1;
                    wcount[left]++;
                    wedges[left][wcount[left]] = v+1;
                    wcount[left]++; 
                     
                }
                else
                {
                    if (right > left)
                        pid = (right*(right+1))/2 - left - 1;
                    else
                        pid = (left*(left+1))/2 - right - 1;
                    bedges[pid][bcount[pid]] = u+1;
                    bcount[pid]++;
                    bedges[pid][bcount[pid]] = v+1;
                    bcount[pid]++;
                }
            
            }
            
        }
        
    }    
    
    
    idx_t ** pedges = (idx_t**)malloc(npes*sizeof(idx_t *));
    idx_t total = 0;
    
    double a = 0.0;
    for(i=0;i < npes;i++)
    {
        a = 1.0  + sqrt(8*i + 1.0);
        right = (idx_t)floor(a*0.5);
        left = right + ((right*(right-1))/2) - i - 1;
        
        total = wcount[left] + wcount[right] + bcount[i];
        pedges[i] = (idx_t*)malloc(total*sizeof(idx_t));
        
        
        //edges within left partition
        for(j=0;j < wcount[left];j++)
        {
            pedges[i][pcount[i]] = wedges[left][j];
            pcount[i]++;
        }
        
        //edges within right partition
        for(j=0;j < wcount[right];j++)
        {
            pedges[i][pcount[i]] = wedges[right][j];
            pcount[i]++;
        }
        
        //edges between left and right partition that get mapped to proc_i
        for(j=0;j < bcount[i];j++)
        {
            pedges[i][pcount[i]] = bedges[i][j];
            pcount[i]++;
        }
        
        
    }

    return pedges;
}

idx_t * exchange_overlaps(idx_t ** pedges, idx_t * pcount,idx_t * recv_count,MPI_Comm comm)
{

    /*************************************************************************
    * This function exchanges partials overlaps.
    * Each process will receive partial edge list from different processes
    **************************************************************************/
    
    int mype, npes,i;
    idx_t offset;//,msize;
    MPI_Comm_size(comm, &npes);
    MPI_Comm_rank(comm, &mype);
    
    MPI_Request *requests;
    MPI_Status  * status;
    requests = (MPI_Request*)malloc((npes * 2)*sizeof(MPI_Request));
    status   = (MPI_Status*)malloc((npes * 2)*sizeof(MPI_Status));
    
    
    
    //Communicate amount to be received
    
    MPI_Alltoall((void *)pcount, 1, IDX_T,(void *)recv_count, 1, IDX_T,MPI_COMM_WORLD);
    
    
    //Create recv buffer based on the counts received from all processes.
    idx_t total_edges_to_recv = 0;
    for(i = 0; i < npes;i++)
    {
        if (i != mype)
            total_edges_to_recv += recv_count[i];
    }
    idx_t * recv_edges = (idx_t *)malloc(total_edges_to_recv*sizeof(idx_t));
    offset = 0;
    for(i=0;i < npes;i++)
    {
        requests[i]=MPI_REQUEST_NULL;
        if (i != mype && recv_count[i] > 0)
        {
            MPI_Irecv((void *)(recv_edges+offset),recv_count[i],IDX_T,i,1,MPI_COMM_WORLD,&(requests[i]));
            offset += recv_count[i];
        }
    }
    
    
    for(i=0;i < npes;i++)
    {
        requests[i+npes]=MPI_REQUEST_NULL;
        if (i != mype && pcount[i] > 0)
            MPI_Isend((void *)pedges[i],pcount[i],IDX_T,i,1,MPI_COMM_WORLD,&(requests[i+npes]));
    }
     
    int err = MPI_Waitall(npes*2, requests, status);
    if (err != MPI_SUCCESS) {
        char err_string[MPI_MAX_ERROR_STRING+1];
        int  i, err_len;
        MPI_Error_string(err, err_string, &err_len);
        printf("Error: MPI_Waitall() at %s line %d (%s)\n",__func__,__LINE__,err_string);
        for (i=0; i<npes*2; i++)
            if (status[i].MPI_ERROR != MPI_SUCCESS) {
                MPI_Error_string(err, err_string, &err_len);
                printf("Error: MPI_Waitall() at %s line %d (status[%d] %s)\n",__func__,__LINE__,i,err_string);
            }
    }
    free(requests);
    free(status);
    

    return recv_edges;
}

idx_t * exchange_root_purity(idx_t* send_buf_root,idx_t num_vertices,idx_t * message_size,idx_t * offset,MPI_Comm comm)
{
    /*
    Vertices are divided block wise so resolution can be faster
    Process communicate community_id for vertices as vertices can be duplicated
    */
    int mype, npes;
    idx_t i,recv_offset,send_offset;
    
    MPI_Comm_size(comm, &npes);
    MPI_Comm_rank(comm, &mype);
    
    idx_t * recv_buf_root =  (idx_t *)calloc((npes-1)*message_size[mype]*2,sizeof(idx_t)); 
    MPI_Request *requests;
    MPI_Status  * status;
    requests = (MPI_Request*)malloc((npes * 2)*sizeof(MPI_Request));
    status   = (MPI_Status*)malloc((npes * 2)*sizeof(MPI_Status));
    recv_offset = 0;
    send_offset = 0;
    
    for(i=0;i < npes;i++)
    {
        requests[i]=MPI_REQUEST_NULL;
        if (i != mype)
        {
            MPI_Irecv((void *)(recv_buf_root+recv_offset),2*message_size[mype],IDX_T,i,1,MPI_COMM_WORLD,&(requests[i]));
            recv_offset += 2*message_size[mype];
        }
    }
    
    
    for(i=0;i < npes;i++)
    {
        requests[i+npes]=MPI_REQUEST_NULL;
        if (i != mype){
            send_offset = 2*offset[i];
            MPI_Isend((void *)(send_buf_root+send_offset),2*message_size[i],IDX_T,i,1,MPI_COMM_WORLD,&(requests[i+npes]));
        }
    }
    int err = MPI_Waitall(npes*2, requests, status);
    if (err != MPI_SUCCESS) {
        char err_string[MPI_MAX_ERROR_STRING+1];
        int  i, err_len;
        MPI_Error_string(err, err_string, &err_len);
        printf("Error: MPI_Waitall() at %s line %d (%s)\n",__func__,__LINE__,err_string);
        for (i=0; i<npes*2; i++)
            if (status[i].MPI_ERROR != MPI_SUCCESS) {
                MPI_Error_string(err, err_string, &err_len);
                printf("Error: MPI_Waitall() at %s line %d (status[%d] %s)\n",__func__,__LINE__,i,err_string);
            }
    }
    free(requests);
    free(status);
    
    return recv_buf_root;
}

idx_t max_root(idx_t a, idx_t b)
{
    if (a >= b)
        return a;
    else
        return b;
}
idx_t * resolve(idx_t * recv_root_purity, idx_t * send_root_purity,idx_t* message_size,idx_t*offset,idx_t npes,idx_t mype)
{
    /*
    Resolve the community for each vertex using heuristic
    Each process finalizes vertices it is in charge of after receiving root from other process
    Resolution is done heuristically. Higher the din value for vertex for a community implies 
    stronger affinity to that community. In case of tie, assign either local or remote comm id
    by chosing max(local_root, remote root)
    */
    idx_t i,v, v_start,v_end,local_root,remote_root,din,din2; 
    
    v_start = 0;///offset[mype] + 1;
    v_end   = v_start + message_size[mype];
    v       = v_start;

    idx_t num_recv_vertices = (npes-1)*message_size[mype]*2;
    idx_t * final_root_purity =  (idx_t *)calloc(2*message_size[mype],sizeof(idx_t));
    for(i=0;i < num_recv_vertices;i+=2)
    {
        if (v == v_end){
            v = v_start;
            
        }
        local_root = send_root_purity[2*(v+offset[mype])];
        din        = send_root_purity[2*(v+offset[mype])+1];
        remote_root = recv_root_purity[i];
        din2 = recv_root_purity[i+1];
        
        
        if(remote_root == 0)
        {
            if(local_root != 0)
            {
                final_root_purity[2*v] = local_root;
                final_root_purity[2*v+1] = din; 
            }
            v++;
            continue;
        }
        
        if (local_root == remote_root)
        {
            final_root_purity[2*v] = local_root;
            final_root_purity[2*v+1] = max_root(din,din2); 
            send_root_purity[2*(v+offset[mype])+1] = max_root(din,din2);
        }
        else{
            if(local_root == 0)
            { 
                final_root_purity[2*v] = remote_root;
                send_root_purity[2*(v+offset[mype])] = remote_root;
                final_root_purity[2*v+1] = din2;
                send_root_purity[2*(v+offset[mype])+1] = din2;
            }
            else if(din2 == din){
                final_root_purity[2*v] = max_root(local_root,remote_root);
                send_root_purity[2*(v+offset[mype])] = max_root(local_root,remote_root);
                final_root_purity[2*v + 1] = din; 
                
            }
            else if(din2 > din){
                final_root_purity[2*v] = remote_root;
                send_root_purity[2*(v+offset[mype])] = remote_root;
                final_root_purity[2*v + 1] = din2;
                send_root_purity[2*(v+offset[mype])+1] = din2; 
            }
            else{
                final_root_purity[2*v] = local_root;
                final_root_purity[2*v + 1] = din;
                
            }
        }
        v++;
    }

    return final_root_purity;
}


                
idx_t * gather_root(idx_t * message_size, idx_t * offset,idx_t * final_root_purity,idx_t num_vertices,MPI_Comm comm)
{
    /*
    Gather the community ids (stored in root variable from all processes and send it to proc 0)
    Required to get final community as the solution is distributed.
    */
    int mype, npes,i;
    
    MPI_Comm_size(comm, &npes);
    MPI_Comm_rank(comm, &mype);
    
    idx_t * recv_data;

    if(mype == 0)
    {
        //recv buffer only matters at root process
        recv_data = (idx_t *)calloc(num_vertices,sizeof(idx_t));
    }
    
    //all processes allocate their own send buffers
    idx_t * send_data = (idx_t *)calloc(message_size[mype],sizeof(idx_t));
    

    //all processes populate the send buffers
    idx_t v = offset[mype] + 1;
    idx_t j = 0;
    for(i=0;i < 2*message_size[mype]; i+=2)
    {
        if (final_root_purity[i] != 0)
            send_data[j] = final_root_purity[i];
        else
            send_data[j] = v;
        v++;
        j++;
    }
    //p0 process gathers data from all processes
    MPI_Gatherv(send_data, message_size[mype], IDX_T, recv_data, message_size, offset, IDX_T, 0, MPI_COMM_WORLD);

    return recv_data;
}

/*************************************************************************
* This function returns root community id
**************************************************************************/
idx_t find_comm_root(idx_t x,idx_t * root)
{
    //printf("In find root\n");
    idx_t v =x+1;
    while(v != root[v-1]){
        //printf("%"PRIDX" %"PRIDX"\n",v,root[v-1]);
        v = root[v-1];
    }
    return v;
}

void write_community(char *gname, idx_t num_vertices, idx_t * final_root, MPI_Comm comm)
{
    
    int npes, mype;

    MPI_Comm_size(comm, &npes);
    MPI_Comm_rank(comm, &mype);
    
    if (mype == 0) 
    {
        idx_t v;
        FILE *fpin;
        char partfile[256];
        sprintf(partfile, "%s_comm%"PRIDX".txt", gname,npes);
        if ((fpin = fopen(partfile, "w")) == NULL) 
        {
            printf("Failed to open file %s\n", partfile);
            MPI_Abort(comm, -1);
        }
        for (v=0; v < num_vertices; v++)
        {
            fprintf(fpin, "%"PRIDX"\t%"PRIDX"\n", v+1,find_comm_root(v,final_root));
        }
        fclose(fpin);
    }
    else{
        return;
    }
}
