/*
 *
 * This file contains routines related to parallel mep algorithm
 *
 * Diana Palsetia
 * 
 * Copyright (C) 2016, Northwestern University
 * See COPYRIGHT notice in top-level directory.
 *
 */

#ifndef H_PMEP
#define H_PMEP

//stdheaders
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <math.h>
#include <stdarg.h>
#include <time.h>
#include <assert.h>

//pmetis header and defines
#include <parmetis.h>
#include <hashtab.h>

#define MAX_STRING_LENGTH 1024
#define MAXNCON         32
#define PMV3_OPTION_DBGLVL        1
#define PMV3_OPTION_SEED        2

//pmetis graph data structure
typedef struct graph_t {
  idx_t gnvtxs, nvtxs, nedges, ncon, nobj;
  idx_t *xadj;     /* Pointers to the locally stored vertices */
  idx_t *vwgt;     /* Vertex weights */
  real_t *nvwgt;   /* Vertex weights */
  idx_t *vsize;    /* Vertex size */
  idx_t *adjncy;   /* Array that stores the adjacency lists of nvtxs */
  idx_t *adjwgt;   /* Array that stores the weights of the adjacency lists */
  idx_t *vtxdist;  /* Distribution of vertices */
  idx_t *home;     /* The initial partition of the vertex */

  /* used for not freeing application supplied arrays */
  idx_t free_vwgt;
  idx_t free_adjwgt;
  idx_t free_vsize;

  /* Coarsening structures */
  idx_t *match;
  idx_t *cmap;

  /* Used during initial partitioning */
  idx_t *label;

  /* Communication/Setup parameters */
  idx_t nnbrs;                  /*!< The number of neighboring processors */
  idx_t nrecv;                  /*!< The total number of remote vertices that need to 
                                     be received. nrecv == recvptr[nnbrs] */
  idx_t nsend;                  /*!< The total number of local vertices that need to 
                                     be sent. This corresponds to the communication 
                                     volume of each pe, in the sense that if a vertex 
                                     needs to be sent multiple times, it is accounted 
                                     in nsend. nsend == sendptr[nnbrs] */
  idx_t *peind;                 /*!< Array of size nnbrs storing the neighboring PEs */
  idx_t *sendptr, *sendind;     /*!< CSR format of the vertices that are sent to each
                                     of the neighboring processors */
  idx_t *recvptr, *recvind;     /*!< CSR format of the vertices that are received from
                                     each of the neighboring PEs. */
  idx_t *imap;                  /*!< The inverse map of local to global indices */
  idx_t *pexadj, *peadjncy, 
        *peadjloc;              /*!< CSR format of the PEs each vertex is adjancent to 
                                     along with the location in the sendind of the 
                                     non-local adjancent vertices */

  idx_t nlocal;                 /*!< Number of interior vertices */
  idx_t *lperm;                 /*!< lperm[0:nlocal] points to interior vertices, 
                                     the rest are interface */
  struct graph_t *coarser, *finer;
} graph_t;


struct vertex_t{
        idx_t key;          /* vertex ID */
        idx_t degree;       /* how many neighbors? */
        idx_t *neighbors;   /* neighbor list */
        idx_t temp_num;     /* used for getting neighbors from the edge list... */
        idx_t free;
        idx_t root;         /* root neighbor */
        idx_t cycle;        /* cycle neighbor */

		idx_t comp;
        idx_t comp_max;
        idx_t compactness;  /* only for root vertices */
        float avg_sep;      /* only for root vertices */
        float purity;       /* only for root vertices */
        void *sep;          /* struct htab * */
};

struct mep_graph_t{
        void *adj;                 /* struct htab * */
        void *comm;                /* struct htab * */
        idx_t num_comm;
        idx_t num_vertex;
        idx_t num_edge;
        idx_t *vertices;           /* sorted by degree */
        struct vertex_t *vtable;
};

int build_free_list(struct mep_graph_t *graph);
void delete_graph(struct mep_graph_t *graph);
int c_serial_version_main(int argc, char **argv);
void cleanup_table(struct htab *table);

int community_grow(struct mep_graph_t *graph);
int community_merge(struct mep_graph_t *graph, idx_t *loops);
int calculate_purity_values(struct mep_graph_t *graph);
idx_t find(struct mep_graph_t *graph, struct vertex_t *vertex);
int printout(char *filename, struct mep_graph_t *graph, idx_t loops);
int find_pure_community(struct mep_graph_t *graph, struct vertex_t *vertex, idx_t *target, unsigned int *found);
int find_community2(struct mep_graph_t *graph, struct vertex_t *vertex, idx_t *target, idx_t *found, idx_t *comp_max);
int find_pure_community_new(struct mep_graph_t *graph, struct vertex_t *vertex, idx_t *target, unsigned int *found);

//pmep function prototypes
idx_t get_pid(idx_t * vtxdist,idx_t v,idx_t npes);
void exchange_partitions(graph_t *graph, idx_t *part, MPI_Comm comm,
     idx_t *recv_part);
void exchange_partitions2_compute(idx_t *dp_count, idx_t **dpid,
     graph_t *graph, idx_t *part, htab *dups,MPI_Comm comm);
htab* exchange_partitions2_comm(idx_t *dp_count, idx_t **dpid,
     graph_t *graph, idx_t *part, MPI_Comm comm);
void create_overlaps(graph_t *graph, idx_t nparts, idx_t *recv_part,
     idx_t **pedges, idx_t *count, MPI_Comm comm);
idx_t** create_overlaps2(graph_t *graph, idx_t nparts, htab *new_part,
     idx_t **bedges, idx_t *bcount, idx_t **wedges, idx_t *wcount,
     idx_t *pcount, MPI_Comm comm);
idx_t* exchange_overlaps(idx_t **pedges, idx_t *pcount,idx_t *recv_count,
     MPI_Comm comm);
idx_t* exchange_root_purity(idx_t *send_buf_root, idx_t num_vertices,
     idx_t *message_size, idx_t *offset, MPI_Comm comm);
idx_t max_root(idx_t a, idx_t b);
idx_t max_din(idx_t a, idx_t b);
idx_t* resolve(idx_t *recv_root_purity, idx_t *send_root_purity,
     idx_t *message_size, idx_t *offset, idx_t npes, idx_t mype);
idx_t* gather_root(idx_t *message_size, idx_t *offset,
     idx_t *final_root_purity, idx_t num_vertices, MPI_Comm comm);
idx_t find_comm_root(idx_t x,idx_t * root);
void write_community(char *gname, idx_t num_vertices, idx_t *final_root,
     MPI_Comm comm);
int create_graph(idx_t num_edges, idx_t *edges, idx_t direction, struct mep_graph_t *pgraph);
void ParallelReadGraph2(char *fname_bin, char *fname_off, idx_t *gnvtxs,
     idx_t *gnedges, idx_t *nvtxs, idx_t *nedges, idx_t **vtxdist,
     idx_t **xadj, idx_t **adjncy, MPI_Comm comm);
int countingSort(idx_t *a, idx_t *b, idx_t num_items, idx_t direction);
int init_data_with_file(char *filename, idx_t type, idx_t *direction, idx_t *num_edges, idx_t **edges);

#endif
